# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).


from psychopy import visual
import gevent
import numpy as np
from collections import OrderedDict
from .....util import convertCamelToSnake, updateSettings
from .... import DeviceEvent, Computer
from .....constants import EventConstants
from psychopy.iohub.devices.keyboard import KeyboardInputEvent

from .....errors import print2err, printExceptionDetailsToStdErr

currentTime = Computer.getTime


class TobiiPsychopyCalibrationGraphics:
    IOHUB_HEARTBEAT_INTERVAL = 0.050  # seconds between forced run through of micro threads, since one is blocking on camera setup.
    CALIBRATION_POINT_LIST = [(0.5, 0.5), (0.1, 0.1), (0.9, 0.1), (0.9, 0.9), (0.1, 0.9), (0.5, 0.5)]

    TEXT_POS = [0, 0]
    TEXT_HEIGHT = 36
    _keyboard_key_index = KeyboardInputEvent.CLASS_ATTRIBUTE_NAMES.index('key')

    def __init__(self, eyetrackerInterface, calibration_args={}):
        self._eyetrackerinterface = eyetrackerInterface
        self._tobii = eyetrackerInterface._tobii
        self.screenSize = eyetrackerInterface._display_device.getPixelResolution()
        self.width = self.screenSize[0]
        self.height = self.screenSize[1]
        self._ioKeyboard = None
        self._msg_queue = []
        self._lastCalibrationOK = False
        display = self._eyetrackerinterface._display_device

        self._device_config = self._eyetrackerinterface.getConfiguration()
        updateSettings(self._device_config.get('calibration'), calibration_args)
        self._calibration_args = self._device_config.get('calibration')
        #print2err("self._calibration_args:", self._calibration_args)
        
        unit_type = self.getCalibSetting('unit_type')
        if unit_type is None:
            unit_type = display.getCoordinateType()
            self._calibration_args['unit_type'] = unit_type
        color_type = self.getCalibSetting('color_type')
        if color_type is None:
            color_type = display.getColorSpace()
            self._calibration_args['color_type'] = color_type

        calibration_methods = dict(THREE_POINTS=3,
                                   FIVE_POINTS=5,
                                   NINE_POINTS=9,
                                   THIRTEEN_POINTS=13)

        cal_type = self.getCalibSetting('type')

        if cal_type in calibration_methods:
            num_points = calibration_methods[cal_type]

            if num_points == 3:
                TobiiPsychopyCalibrationGraphics.CALIBRATION_POINT_LIST = [
                    (0.5, 0.1), (0.1, 0.9), (0.9, 0.9)]
            elif num_points == 5:
                TobiiPsychopyCalibrationGraphics.CALIBRATION_POINT_LIST = [(0.5, 0.5),
                                                                           (0.1, 0.1),
                                                                           (0.9, 0.1),
                                                                           (0.9, 0.9),
                                                                           (0.1, 0.9)]
            elif num_points == 9:
                TobiiPsychopyCalibrationGraphics.CALIBRATION_POINT_LIST = [(0.5, 0.5),
                                                                           (0.1,
                                                                            0.5),
                                                                           (0.9,
                                                                            0.5),
                                                                           (0.1,
                                                                            0.1),
                                                                           (0.5,
                                                                            0.1),
                                                                           (0.9,
                                                                            0.1),
                                                                           (0.9,
                                                                            0.9),
                                                                           (0.5,
                                                                            0.9),
                                                                           (0.1,
                                                                            0.9)]
            elif num_points == 13:
                TobiiPsychopyCalibrationGraphics.CALIBRATION_POINT_LIST = [(0.5, 0.5),
                                                                               (0.1, 0.5),
                                                                               (0.9, 0.5),
                                                                               (0.1, 0.1),
                                                                               (0.5, 0.1),
                                                                               (0.9, 0.1),
                                                                               (0.9, 0.9),
                                                                               (0.5, 0.9),
                                                                               (0.1, 0.9),
                                                                               (0.25, 0.25),
                                                                               (0.25, 0.75),
                                                                               (0.75, 0.75),
                                                                               (0.75, 0.25)
                                                                               ]

        self.window = visual.Window(
            self.screenSize,
            monitor=display.getPsychopyMonitorName(),
            units=unit_type,
            fullscr=True,
            allowGUI=False,
            screen=display.getIndex(),
            color=self.getCalibSetting(['screen_background_color']),
            colorSpace=color_type)
        self.window.flip(clearBuffer=True)

        self._createStim()
        self._registerEventMonitors()
        self._lastMsgPumpTime = currentTime()

        self.clearAllEventBuffers()

    def getCalibSetting(self, setting):
        if isinstance(setting, str):
            setting = [setting, ]
        calibration_args = self._calibration_args
        if setting:
            for s in setting[:-1]:
                calibration_args = calibration_args.get(s)
            return calibration_args.get(setting[-1])

    def clearAllEventBuffers(self):
        self._eyetrackerinterface._iohub_server.eventBuffer.clear()
        for d in self._eyetrackerinterface._iohub_server.devices:
            d.clearEvents()

    def _registerEventMonitors(self):
        kbDevice = None
        if self._eyetrackerinterface._iohub_server:
            for dev in self._eyetrackerinterface._iohub_server.devices:
                if dev.__class__.__name__ == 'Keyboard':
                    kbDevice = dev

        if kbDevice:
            eventIDs = []
            for event_class_name in kbDevice.__class__.EVENT_CLASS_NAMES:
                eventIDs.append(
                    getattr(
                        EventConstants,
                        convertCamelToSnake(
                            event_class_name[
                            :-5],
                            False)))

            self._ioKeyboard = kbDevice
            self._ioKeyboard._addEventListener(self, eventIDs)
        else:
            print2err(
                'Warning: Tobii Cal GFX could not connect to Keyboard device for events.')

    def _unregisterEventMonitors(self):
        if self._ioKeyboard:
            self._ioKeyboard._removeEventListener(self)

    def _handleEvent(self, event):
        event_type_index = DeviceEvent.EVENT_TYPE_ID_INDEX
        if event[event_type_index] == EventConstants.KEYBOARD_PRESS:
            ek = event[self._keyboard_key_index]
            if isinstance(ek, bytes):
                ek = ek.decode('utf-8')
            if ek == ' ' or ek == 'space':
                self._msg_queue.append('SPACE_KEY_ACTION')
                self.clearAllEventBuffers()
            elif ek == 'escape':
                self._msg_queue.append('QUIT')
                self.clearAllEventBuffers()

    def MsgPump(self):
        # keep the psychopy window happy ;)
        if currentTime() - self._lastMsgPumpTime > self.IOHUB_HEARTBEAT_INTERVAL:
            # try to keep ioHub from being blocked. ;(
            if self._eyetrackerinterface._iohub_server:
                for dm in self._eyetrackerinterface._iohub_server.deviceMonitors:
                    dm.device._poll()
                self._eyetrackerinterface._iohub_server.processDeviceEvents()
            self._lastMsgPumpTime = currentTime()

    def getNextMsg(self):
        if len(self._msg_queue) > 0:
            msg = self._msg_queue[0]
            self._msg_queue = self._msg_queue[1:]
            return msg

    def _createStim(self):
        """
            outer_diameter: 35
            outer_stroke_width: 5
            outer_fill_color: [255,255,255]
            outer_line_color: [255,255,255]
            inner_diameter: 5
            inner_stroke_width: 0
            inner_color: [0,0,0]
            inner_fill_color: [0,0,0]
            inner_line_color: [0,0,0]
            calibration_prefs=self._eyetrackerinterface.getConfiguration()['calibration']['target_attributes']
        """
        color_type = self.getCalibSetting('color_type')
        unit_type = self.getCalibSetting('unit_type')

        self.calibrationPoint = visual.TargetStim(
            self.window, name="CP", style="circles",
            radius=self.getCalibSetting(['target_attributes', 'outer_diameter']) / 2.0,
            fillColor=self.getCalibSetting(['target_attributes', 'outer_fill_color']),
            borderColor=self.getCalibSetting(['target_attributes', 'outer_line_color']),
            lineWidth=self.getCalibSetting(['target_attributes', 'outer_stroke_width']),
            innerRadius=self.getCalibSetting(['target_attributes', 'inner_diameter']) / 2.0,
            innerFillColor=self.getCalibSetting(['target_attributes', 'inner_fill_color']),
            innerBorderColor=self.getCalibSetting(['target_attributes', 'inner_line_color']),
            innerLineWidth=self.getCalibSetting(['target_attributes', 'inner_stroke_width']),
            pos=(0, 0),
            units=unit_type,
            colorSpace=color_type,
            autoLog=False
        )
        self.calibrationPointINNER = self.calibrationPoint.inner
        self.calibrationPointOUTER = self.calibrationPoint.outer

        tctype = color_type
        tcolor = self.getCalibSetting(['text_color'])
        if tcolor is None:
            # If no calibration text color provided, base it on the window background color
            from psychopy.iohub.util import complement
            sbcolor = self.getCalibSetting(['screen_background_color'])
            if sbcolor is None:
                sbcolor = self.window.color
            from psychopy.colors import Color
            tcolor_obj = Color(sbcolor, color_type)
            tcolor = complement(*tcolor_obj.rgb255)
            tctype = 'rgb255'

        instuction_text = 'Press SPACE to Start Calibration; ESCAPE to Exit.'
        self.textLineStim = visual.TextStim(self.window,
                                            text=instuction_text,
                                            pos=self.TEXT_POS,
                                            height=self.TEXT_HEIGHT,
                                            color=tcolor,
                                            colorSpace=tctype,
                                            units='pix',
                                            wrapWidth=self.width * 0.9)

        # create Tobii eye position feedback graphics
        #
        sw, sh = self.screenSize
        self.hbox_bar_length = hbox_bar_length = sw / 4
        hbox_bar_height = 6
        marker_diameter = 7
        self.marker_heights = (-sh / 2.0 * .7, -sh / 2.0 * .75, -sh /
                               2.0 * .8, -sh / 2.0 * .7, -sh / 2.0 * .75, -sh / 2.0 * .8)

        bar_vertices = ([-hbox_bar_length / 2, -hbox_bar_height / 2], [hbox_bar_length / 2, -hbox_bar_height / 2],
                        [hbox_bar_length / 2, hbox_bar_height / 2], [-hbox_bar_length / 2, hbox_bar_height / 2])

        self.feedback_resources = OrderedDict()

        self.feedback_resources['hbox_bar_x'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='Firebrick',
            vertices=bar_vertices,
            units='pix',
            pos=(
                0,
                self.marker_heights[0]))
        self.feedback_resources['hbox_bar_y'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='DarkSlateGray',
            vertices=bar_vertices,
            units='pix',
            pos=(
                0,
                self.marker_heights[1]))
        self.feedback_resources['hbox_bar_z'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='GoldenRod',
            vertices=bar_vertices,
            units='pix',
            pos=(
                0,
                self.marker_heights[2]))

        marker_vertices = [-marker_diameter, 0], [0, marker_diameter], [marker_diameter, 0], [0, -marker_diameter]
        self.feedback_resources['left_hbox_marker_x'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='Black',
            vertices=marker_vertices,
            units='pix',
            pos=(
                0,
                self.marker_heights[0]))
        self.feedback_resources['left_hbox_marker_y'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='Black',
            units='pix',
            vertices=marker_vertices,
            pos=(
                0,
                self.marker_heights[1]))
        self.feedback_resources['left_hbox_marker_z'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='Black',
            units='pix',
            vertices=marker_vertices,
            pos=(
                0,
                self.marker_heights[2]))
        self.feedback_resources['right_hbox_marker_x'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='DimGray',
            units='pix',
            vertices=marker_vertices,
            pos=(
                0,
                self.marker_heights[0]))
        self.feedback_resources['right_hbox_marker_y'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='DimGray',
            units='pix',
            vertices=marker_vertices,
            pos=(
                0,
                self.marker_heights[1]))
        self.feedback_resources['right_hbox_marker_z'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='DimGray',
            units='pix',
            vertices=marker_vertices,
            pos=(
                0,
                self.marker_heights[2]))

    def runCalibration(self):
        """
        Performs a Tobii calibration routine.
        """

        self._lastCalibrationOK = False
        calibration_sequence_completed = False

        instuction_text = 'Press SPACE to Start Calibration; ESCAPE to Exit.'
        continue_calibration = self.showSystemSetupMessageScreen(instuction_text, True)
        if not continue_calibration:
            return False
        auto_pace = self.getCalibSetting('auto_pace')
        pacing_speed = self.getCalibSetting('pacing_speed')
        randomize_points = self.getCalibSetting('randomize')

        movement_velocity = self.getCalibSetting(['target_attributes', 'animate', 'movement_velocity'])
        expansion_speed = self.getCalibSetting(['target_attributes', 'animate', 'expansion_speed'])
        use_deprecated_gfx = movement_velocity or expansion_speed
        if not use_deprecated_gfx:
            # If using calibration option as of 2021.2, set pacing_speed
            # to match target_delay
            pacing_speed = self.getCalibSetting('target_delay')

        cal_target_list = self.CALIBRATION_POINT_LIST[1:]
        if randomize_points is True:
            import random
            random.seed(None)
            random.shuffle(cal_target_list)
        cal_target_list.insert(0, self.CALIBRATION_POINT_LIST[0])

        calibration = self._tobii.newScreenCalibration()
        calibration.enter_calibration_mode()

        i = 0
        _quit = False
        left, top, right, bottom = self._eyetrackerinterface._display_device.getCoordBounds()
        w, h = right - left, top - bottom
        for pt in cal_target_list:
            self.clearAllEventBuffers()
            x, y = left + w * pt[0], bottom + h * (1.0 - pt[1])
            if use_deprecated_gfx:
                self.drawCalibrationTargetDeprecated(i, (x, y))
            else:
                start_time = currentTime()

                # Target animate / delay
                animate_enable = self.getCalibSetting(['target_attributes', 'animate', 'enable'])
                animate_expansion_ratio = self.getCalibSetting(['target_attributes', 'animate', 'expansion_ratio'])
                animate_contract_only = self.getCalibSetting(['target_attributes', 'animate', 'contract_only'])
                target_delay = self.getCalibSetting(['target_delay'])
                target_duration = self.getCalibSetting(['target_duration'])
                while currentTime()-start_time <= target_delay:
                    if animate_enable and i > 0:
                        t = (currentTime()-start_time) / target_delay
                        v1 = cal_target_list[i-1]
                        v2 = pt
                        t = 60.0 * ((1.0 / 10.0) * t ** 5 - (1.0 / 4.0) * t ** 4 + (1.0 / 6.0) * t ** 3)
                        mx, my = ((1.0 - t) * v1[0] + t * v2[0], (1.0 - t) * v1[1] + t * v2[1])
                        moveTo = left + w * mx, bottom + h * (1.0 - my)
                        self.drawCalibrationTarget(moveTo, reset=False)
                    else:
                        self.drawCalibrationTarget((x, y), False)

                    self.getNextMsg()
                    self.MsgPump()

                self.drawCalibrationTarget((x, y))
                start_time = currentTime()

                outer_diameter = self.getCalibSetting(['target_attributes', 'outer_diameter'])
                inner_diameter = self.getCalibSetting(['target_attributes', 'inner_diameter'])
                while currentTime()-start_time <= target_duration:
                    elapsed_time = currentTime()-start_time
                    d = t = None
                    if animate_contract_only:
                        # Change target size from outer diameter to inner diameter over target_duration seconds.
                        t = elapsed_time / target_duration
                        d = outer_diameter - t * (outer_diameter - inner_diameter)
                        self.calibrationPoint.outerRadius = d / 2
                        self.calibrationPoint.draw()
                        self.window.flip(clearBuffer=True)
                    elif animate_expansion_ratio not in [1, 1.0]:
                        if elapsed_time <= target_duration/2:
                            # In expand phase
                            t = elapsed_time / (target_duration/2)
                            d = outer_diameter + t * (outer_diameter*animate_expansion_ratio - outer_diameter)
                        else:
                            # In contract phase
                            t = (elapsed_time-target_duration/2) / (target_duration/2)
                            d = outer_diameter*animate_expansion_ratio - t * (outer_diameter*animate_expansion_ratio - inner_diameter)
                    if d:
                        self.calibrationPoint.outerRadius = d / 2
                        self.calibrationPoint.draw()
                        self.window.flip(clearBuffer=True)

                    self.getNextMsg()
                    self.MsgPump()
                    gevent.sleep(0.001)

            self.clearAllEventBuffers()

            stime = currentTime()
            def waitingForNextTargetTime():
                return True

            if auto_pace is True:
                def waitingForNextTargetTime():
                    if use_deprecated_gfx:
                        return currentTime() - stime < float(pacing_speed)
                    return False

            _quit = False
            while waitingForNextTargetTime():
                msg = self.getNextMsg()
                if msg == 'SPACE_KEY_ACTION':
                    break
                elif msg == 'QUIT':
                    _quit = True
                    break
                self.MsgPump()
                gevent.sleep(0.01)

            calibration.collect_data(pt[0], pt[1])

            if _quit:
                calibration.leave_calibration_mode()
                calibration = None
                break

            self.clearCalibrationWindow()
            self.clearAllEventBuffers()

            i += 1
            if i == len(cal_target_list):
                calibration_sequence_completed = True

        self.clearCalibrationWindow()
        self.clearAllEventBuffers()

        cal_result_dict = None
        if _quit:
            return cal_result_dict

        self._lastCalibrationOK = False
        if calibration:
            if calibration_sequence_completed:
                calibration_result = calibration.compute_and_apply()
                cal_result_dict = dict(status=calibration_result.status)
                cal_result_dict['points']=[]
                for cp in calibration_result.calibration_points:
                    csamples = []
                    for cs in cp.calibration_samples:
                        csamples.append((cs.left_eye.position_on_display_area, cs.left_eye.validity))
                    cal_result_dict['points'].append((cp.position_on_display_area, csamples))
                self._lastCalibrationOK = calibration_result.status == 'calibration_status_success'
            else:
                self._lastCalibrationOK = False
            calibration.leave_calibration_mode()
            calibration = None

        if self._lastCalibrationOK is False:
            instuction_text = 'Calibration Failed. Options: SPACE: Re-run Calibration; ESCAPE: Exit Setup'
            continue_method = self.showSystemSetupMessageScreen(
                instuction_text, True, msg_types=['SPACE_KEY_ACTION', 'QUIT'])
            if continue_method is False:
                return self.runCalibration()
            return cal_result_dict

        instuction_text = "Calibration Passed. PRESS 'SPACE' KEY TO CONTINUE."
        self.showSystemSetupMessageScreen(instuction_text, True, msg_types=['SPACE_KEY_ACTION'])

        return cal_result_dict   

    def clearCalibrationWindow(self):
        self.window.flip(clearBuffer=True)

    def showSystemSetupMessageScreen(self, text_msg='Press SPACE to Start Calibration; ESCAPE to Exit.',
                                     enable_recording=False, msg_types=('SPACE_KEY_ACTION', 'QUIT')):
        if enable_recording is True:
            self._eyetrackerinterface.setRecordingState(True)

        self.clearAllEventBuffers()

        while True:
            self.textLineStim.setText(text_msg)
            event_named_tuples = []
            for e in self._eyetrackerinterface.getEvents(EventConstants.BINOCULAR_EYE_SAMPLE):
                event_named_tuples.append(
                    EventConstants.getClass(EventConstants.BINOCULAR_EYE_SAMPLE).createEventAsNamedTuple(e))
            leye_box_pos, reye_box_pos = self.getHeadBoxPosition(event_named_tuples)
            lx, ly, lz = leye_box_pos
            rx, ry, rz = reye_box_pos
            eye_positions = (lx, ly, lz, rx, ry, rz)
            marker_names = (
                'left_hbox_marker_x',
                'left_hbox_marker_y',
                'left_hbox_marker_z',
                'right_hbox_marker_x',
                'right_hbox_marker_y',
                'right_hbox_marker_z')
            marker_heights = self.marker_heights
            hbox_bar_length = self.hbox_bar_length

            for i, p in enumerate(eye_positions):
                if p is not None:
                    mpoint = hbox_bar_length * p - hbox_bar_length / 2.0, marker_heights[i]
                    self.feedback_resources[marker_names[i]].setPos(mpoint)
                    self.feedback_resources[marker_names[i]].setOpacity(1.0)
                else:
                    self.feedback_resources[marker_names[i]].setOpacity(0.0)

            self.textLineStim.draw()
            [r.draw() for r in self.feedback_resources.values()]
            self.window.flip()

            msg = self.getNextMsg()
            if msg == 'SPACE_KEY_ACTION':
                if enable_recording is True:
                    self._eyetrackerinterface.setRecordingState(False)
                self.clearAllEventBuffers()
                return True
            elif msg == 'QUIT':
                if enable_recording is True:
                    self._eyetrackerinterface.setRecordingState(False)
                self.clearAllEventBuffers()
                return False
            self.MsgPump()
            gevent.sleep()

    def getHeadBoxPosition(self, events):
        # KeyboardInputEvent.CLASS_ATTRIBUTE_NAMES.index('key_id')
        left_eye_cam_x = None
        left_eye_cam_y = None
        left_eye_cam_z = None
        right_eye_cam_x = None
        right_eye_cam_y = None
        right_eye_cam_z = None

        if len(events) == 0:
            return (left_eye_cam_x, left_eye_cam_y, left_eye_cam_z), (right_eye_cam_x, right_eye_cam_y, right_eye_cam_z)

        event = events[-1]
        if abs(event.left_eye_cam_x) != 1.0 and abs(event.left_eye_cam_y) != 1.0:
            left_eye_cam_x = 1.0 - event.left_eye_cam_x
            left_eye_cam_y = event.left_eye_cam_y
        if event.left_eye_cam_z != 0.0:
            left_eye_cam_z = event.left_eye_cam_z
        if abs(event.right_eye_cam_x) != 1.0 and abs(event.right_eye_cam_y) != 1.0:
            right_eye_cam_x = 1.0 - event.right_eye_cam_x
            right_eye_cam_y = event.right_eye_cam_y
        if event.right_eye_cam_z != 0.0:
            right_eye_cam_z = event.right_eye_cam_z
        return (left_eye_cam_x, left_eye_cam_y, left_eye_cam_z), (right_eye_cam_x, right_eye_cam_y, right_eye_cam_z)

    def setTargetDefaults(self):
        """
            outer_diameter: 35
            outer_stroke_width: 5
            outer_fill_color: [255,255,255]
            outer_line_color: [255,255,255]
            inner_diameter: 5
            inner_stroke_width: 0
            inner_color: [0,0,0]
            inner_fill_color: [0,0,0]
            inner_line_color: [0,0,0]
        """
        self.calibrationPointOUTER.radius = self.getCalibSetting(['target_attributes', 'outer_diameter']) / 2.0
        self.calibrationPointOUTER.setLineColor(self.getCalibSetting(['target_attributes', 'outer_line_color']))
        self.calibrationPointOUTER.setFillColor(self.getCalibSetting(['target_attributes', 'outer_fill_color']))
        self.calibrationPointOUTER.lineWidth = int(self.getCalibSetting(['target_attributes', 'outer_stroke_width']))
        self.calibrationPointINNER.radius = self.getCalibSetting(['target_attributes', 'inner_diameter']) / 2.0
        self.calibrationPointINNER.setLineColor(self.getCalibSetting(['target_attributes', 'inner_line_color']))
        self.calibrationPointINNER.setFillColor(self.getCalibSetting(['target_attributes', 'inner_fill_color']))
        self.calibrationPointINNER.lineWidth = int(self.getCalibSetting(['target_attributes', 'inner_stroke_width']))

        self.calibrationPointOUTER.draw()
        self.calibrationPointINNER.draw()
        return self.window.flip(clearBuffer=True)

    def moveTarget(self, start_pt, end_pt, TARG_VELOCITY):
        sx, sy = start_pt
        ex, ey = end_pt
        dist = np.linalg.norm(end_pt - start_pt)
        sec_dur = dist / TARG_VELOCITY
        num_retraces = sec_dur / self._eyetrackerinterface._display_device.getRetraceInterval()
        x_points = np.linspace(sx, ex, num=int(num_retraces))
        y_points = np.linspace(sy, ey, num=int(num_retraces))
        t_points = zip(x_points, y_points)
        for p in t_points:
            self.calibrationPointOUTER.setPos(p)
            self.calibrationPointINNER.setPos(p)
            self.calibrationPointOUTER.draw()
            self.calibrationPointINNER.draw()
            self.window.flip(clearBuffer=True)
        self.setTargetDefaults()

    def expandTarget(self, TARG_RAD_MULTIPLIER, EXPANSION_RATE):
        orad = self.getCalibSetting(['target_attributes', 'outer_diameter']) / 2.0
        self.calibrationPointOUTER.lineWidth = int(self.getCalibSetting(['target_attributes', 'outer_stroke_width']))
        if self.calibrationPointOUTER.lineWidth < 1:
            self.calibrationPointOUTER.lineWidth = 1

        max_osize = orad * TARG_RAD_MULTIPLIER
        if EXPANSION_RATE < 1:
            EXPANSION_RATE = 1.0

        stime = Computer.getTime()
        self.calibrationPoint.outerRadius = orad
        self.calibrationPoint.draw()
        ftime = self.window.flip(clearBuffer=True)
        current_size = self.calibrationPoint.outerRadius
        while current_size < max_osize:
            sec_dur = ftime - stime
            if sec_dur < 0.0:
                sec_dur = 0.0
            stime = ftime
            current_size += sec_dur * EXPANSION_RATE
            self.calibrationPoint.outerRadius = current_size
            self.calibrationPoint.draw()
            ftime = self.window.flip(clearBuffer=True)

    def contractTarget(self, TARG_RAD_MULTIPLIER, EXPANSION_RATE):
        orad = self.getCalibSetting(['target_attributes', 'outer_diameter']) / 2.0
        self.calibrationPointOUTER.lineWidth = int(self.getCalibSetting(['target_attributes', 'outer_stroke_width']))
        if self.calibrationPointOUTER.lineWidth < 1:
            self.calibrationPointOUTER.lineWidth = 1

        max_osize = orad * TARG_RAD_MULTIPLIER
        if EXPANSION_RATE < 1:
            EXPANSION_RATE = 1.0

        stime = Computer.getTime()
        self.calibrationPoint.outerRadius = max_osize
        self.calibrationPoint.draw()
        ftime = self.window.flip(clearBuffer=True)
        current_size = max_osize
        while current_size > orad:
            sec_dur = ftime - stime
            if sec_dur < 0.0:
                sec_dur = 0.0
            stime = ftime
            current_size -= sec_dur * EXPANSION_RATE
            self.calibrationPoint.outerRadius = current_size
            self.calibrationPoint.draw()
            ftime = self.window.flip(clearBuffer=True)

    def drawCalibrationTargetDeprecated(self, target_number, tp):
        """
            outer_diameter: 35
            outer_stroke_width: 5
            outer_fill_color: [255,255,255]
            outer_line_color: [255,255,255]
            inner_diameter: 5
            inner_stroke_width: 0
            inner_color: [0,0,0]
            inner_fill_color: [0,0,0]
            inner_line_color: [0,0,0]
            calibration_prefs=self._eyetrackerinterface.getConfiguration()['calibration']['target_attributes']
        """
        try:
            animate_prefs = self.getCalibSetting(['target_attributes', 'animate'])

            if animate_prefs:
                CONTRACT_ONLY = animate_prefs.get('contract_only', False)
                TARG_VELOCITY = animate_prefs.get('movement_velocity', 300.0)  # 200 pix / sec
                TARG_RAD_MULTIPLIER = animate_prefs.get('expansion_ratio', 3.0)
                EXPANSION_RATE = animate_prefs.get('expansion_speed', 30.0)

                if target_number == 0:
                    # Do first point animation
                    self.calibrationPointOUTER.setPos(tp)
                    self.calibrationPointINNER.setPos(tp)
                    self.setTargetDefaults()
                    if CONTRACT_ONLY is False:
                        self.expandTarget(TARG_RAD_MULTIPLIER, EXPANSION_RATE)
                    self.contractTarget(TARG_RAD_MULTIPLIER, EXPANSION_RATE)
                else:
                    # Move from current point to new point
                    # then do point animation
                    spos = self.calibrationPointOUTER.pos
                    # self.calibrationPointOUTER.setPos(tp)
                    # self.calibrationPointINNER.setPos(tp)
                    if TARG_VELOCITY > 0.0:
                        self.moveTarget(spos, tp, TARG_VELOCITY)
                    else:
                        self.calibrationPointOUTER.setPos(tp)
                        self.calibrationPointINNER.setPos(tp)
                    self.setTargetDefaults()
                    if CONTRACT_ONLY is False:
                        self.expandTarget(TARG_RAD_MULTIPLIER, EXPANSION_RATE)
                    self.contractTarget(TARG_RAD_MULTIPLIER, EXPANSION_RATE)
            else:
                self.calibrationPointOUTER.setPos(tp)
                self.calibrationPointINNER.setPos(tp)
                self.setTargetDefaults()

        except Exception:
            printExceptionDetailsToStdErr()

    def drawDefaultTarget(self):
        self.calibrationPointOUTER.radius = self.getCalibSetting(['target_attributes', 'outer_diameter']) / 2.0
        self.calibrationPointOUTER.setLineColor(self.getCalibSetting(['target_attributes', 'outer_line_color']))
        self.calibrationPointOUTER.setFillColor(self.getCalibSetting(['target_attributes', 'outer_fill_color']))
        self.calibrationPointOUTER.lineWidth = int(self.getCalibSetting(['target_attributes', 'outer_stroke_width']))
        self.calibrationPointINNER.radius = self.getCalibSetting(['target_attributes', 'inner_diameter']) / 2.0
        self.calibrationPointINNER.setLineColor(self.getCalibSetting(['target_attributes', 'inner_line_color']))
        self.calibrationPointINNER.setFillColor(self.getCalibSetting(['target_attributes', 'inner_fill_color']))
        self.calibrationPointINNER.lineWidth = int(self.getCalibSetting(['target_attributes', 'inner_stroke_width']))

        self.calibrationPoint.draw()
        return self.window.flip(clearBuffer=True)

    def drawCalibrationTarget(self, tp, reset=True):
        self.calibrationPoint.pos = tp
        if reset:
            return self.drawDefaultTarget()
        else:
            self.calibrationPoint.draw()
            return self.window.flip(clearBuffer=True)



from pathlib import Path

from . import _TestBaseComponentsMixin, _TestDisabledMixin
from .test_base_components import _find_global_resource_in_js_experiment
from psychopy.experiment.components.settings import SettingsComponent
from psychopy import experiment
from ...utils import TESTS_DATA_PATH


class TestSettingsComponent(_TestBaseComponentsMixin):
    def test_unhandled_resources_js(self):
        """
        Check that resources not otherwise handled are present at the start of the experiment
        """
        cases = [
            # Resource not handled, no loop present
            {'exp': "unhandled_noloop",
             'seek': ['blue.png'],
             'avoid': ['white.png', 'yellow.png', 'groups.csv', 'groupA.csv', 'groupB.csv']},
            # Resource not handled, loop defined by string
            {'exp': "unhandled_strloop",
             'seek': ['blue.png', 'white.png', 'groupA.csv'],
             'avoid': ['yellow.png', 'groupB.csv', 'groups.csv']},
            # Resource not handled, loop defined by constructed string
            {'exp': "unhandled_constrloop",
             'seek': ['blue.png', 'white.png', 'yellow.png', 'groupA.csv', 'groupB.csv', 'groups.csv'],
             'avoid': []},
            # Resource not handled, loop defined by constructed string from loop
            {'exp': "unhandled_recurloop",
             'seek': ['blue.png', 'white.png', 'yellow.png', 'groupA.csv', 'groupB.csv', 'groups.csv'],
             'avoid': []},
        ]

        exp = experiment.Experiment()
        for case in cases:
            # Load experiment
            exp.loadFromXML(Path(TESTS_DATA_PATH) / "test_get_resources" / (case['exp'] + ".psyexp"))
            # Write to JS
            script = exp.writeScript(target="PsychoJS")
            # Check that all "seek" phrases are included
            for phrase in case['seek']:
                assert _find_global_resource_in_js_experiment(script, phrase), (
                    f"'{phrase}' was not found in resources for {case['exp']}.psyexp"
                )
            # Check that all "avoid" phrases are excluded
            for phrase in case['avoid']:
                assert not _find_global_resource_in_js_experiment(script, phrase), (
                    f"'{phrase}' was found in resources for {case['exp']}.psyexp"
                )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Classes and functions for managing audio devices.
"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

__all__ = [
    'AudioDeviceInfo',
    'AudioDeviceStatus',
    'NULL_AUDIO_DEVICE',
    'NULL_AUDIO_DEVICE_STATUS',
    'sampleRateQualityLevels',
    'latencyClassLevels',
    'runModeLevels',
    'SAMPLE_RATE_8kHz',
    'SAMPLE_RATE_TELCOM_QUALITY',
    'SAMPLE_RATE_16kHz',
    'SAMPLE_RATE_VOIP_QUALITY',
    'SAMPLE_RATE_VOICE_QUALITY',
    'SAMPLE_RATE_44p1kHz',
    'SAMPLE_RATE_CD_QUALITY',
    'SAMPLE_RATE_48kHz',
    'SAMPLE_RATE_DVD_QUALITY',
    'SAMPLE_RATE_96kHz',
    'SAMPLE_RATE_192kHz',
    'AUDIO_PTB_LATENCY_CLASS0',
    'AUDIO_PTB_LATENCY_CLASS1',
    'AUDIO_PTB_LATENCY_CLASS2',
    'AUDIO_PTB_LATENCY_CLASS3',
    'AUDIO_PTB_LATENCY_CLASS4',
    'AUDIO_PTB_LATENCY_CLASS_DONT_CARE',
    'AUDIO_PTB_LATENCY_CLASS_SHARE',
    'AUDIO_PTB_LATENCY_CLASS_EXCLUSIVE',
    'AUDIO_PTB_LATENCY_CLASS_AGGRESSIVE',
    'AUDIO_PTB_LATENCY_CLASS_CRITICAL',
    'AUDIO_PTB_RUN_MODE0',
    'AUDIO_PTB_RUN_MODE1',
    'AUDIO_PTB_RUN_MODE_STANDBY',
    'AUDIO_PTB_RUN_MODE_KEEP_HOT',
    'AUDIO_LIBRARY_PTB'
]

from psychopy.tools.audiotools import *

# audio library identifiers
AUDIO_LIBRARY_PTB = 'ptb'  # PsychPortAudio from Psychtoolbox

# Quality levels as strings and values. Used internally by the PsychoPy UI for
# dropdowns and preferences. Persons using PsychoPy as a library would typically
# use constants `SAMPLE_RATE_*` instead of looking up values in here.
#
# For voice recording applications, the recommended sample rate is `Voice`
# (16kHz) and should appear as the default option in preferences and UI
# dropdowns.
#
sampleRateQualityLevels = {
    0: (SAMPLE_RATE_8kHz, 'Telephone/Two-way radio (8kHz)'),
    1: (SAMPLE_RATE_16kHz, 'Voice (16kHz)'),  # <<< recommended for voice
    2: (SAMPLE_RATE_44p1kHz, 'CD Audio (44.1kHz)'),
    3: (SAMPLE_RATE_48kHz, 'DVD Audio (48kHz)'),  # <<< usually system default
    4: (SAMPLE_RATE_96kHz, 'High-Def (96kHz)'),
    5: (SAMPLE_RATE_192kHz, 'Ultra High-Def (192kHz)')
}

# Latency classes for the PsychPortAudio backend. These are used to set how
# aggressive PsychPortAudio should be at minimizing sound latency and getting
# precise timing. Exclusive mode `AUDIO_PTB_LATENCY_CLASS2` is usually used
# for the best timing and maximum compatibility.
#
AUDIO_PTB_LATENCY_CLASS0 = AUDIO_PTB_LATENCY_CLASS_DONT_CARE = 0
AUDIO_PTB_LATENCY_CLASS1 = AUDIO_PTB_LATENCY_CLASS_SHARE = 1
AUDIO_PTB_LATENCY_CLASS2 = AUDIO_PTB_LATENCY_CLASS_EXCLUSIVE = 2
AUDIO_PTB_LATENCY_CLASS3 = AUDIO_PTB_LATENCY_CLASS_AGGRESSIVE = 3
AUDIO_PTB_LATENCY_CLASS4 = AUDIO_PTB_LATENCY_CLASS_CRITICAL = 4

# used for GUI dropdowns
latencyClassLevels = {
    0: (AUDIO_PTB_LATENCY_CLASS0, 'Latency not important'),
    1: (AUDIO_PTB_LATENCY_CLASS1, 'Share low-latency driver'),
    2: (AUDIO_PTB_LATENCY_CLASS2, 'Exclusive low-latency'),  # <<< default
    3: (AUDIO_PTB_LATENCY_CLASS3, 'Aggressive low-latency'),
    4: (AUDIO_PTB_LATENCY_CLASS4, 'Latency critical'),
}

# Run modes for the PsychPortAudio backend.
AUDIO_PTB_RUN_MODE0 = AUDIO_PTB_RUN_MODE_STANDBY = 0
AUDIO_PTB_RUN_MODE1 = AUDIO_PTB_RUN_MODE_KEEP_HOT = 1

runModeLevels = {
    0: (AUDIO_PTB_RUN_MODE0, 'Standby (low resource use, higher latency)'),
    1: (AUDIO_PTB_RUN_MODE1, 'Keep hot (higher resource use, low latency)')
}


class AudioDeviceInfo:
    """Descriptor for an audio device (playback or recording) on this system.

    Properties associated with this class provide information about a specific
    audio playback or recording device. An object can be then passed to
    :class:`~psychopy.sound._microphone.Microphone` to open a stream using the
    device described by the object.

    This class is usually instanced only by calling
    :meth:`~psychopy.sound._microphone.Microphone.getDevices()`. Users should
    avoid creating instances of this class themselves unless they have good
    reason to.

    Parameters
    ----------
    deviceIndex : int
        Enumerated index of the audio device. This number is specific to the
        engine used for audio.
    deviceName : str
        Human-readable name of the device.
    hostAPIName : str
        Human-readable name of the host API used for audio.
    outputChannels : int
        Number of output channels.
    outputLatency : tuple
        Low (`float`) and high (`float`) output latency in milliseconds.
    inputChannels : int
        Number of input channels.
    inputLatency : tuple
        Low (`float`) and high (`float`) input latency in milliseconds.
    defaultSampleRate : int
        Default sample rate for the device in Hertz (Hz).
    audioLib : str
        Audio library that queried device information used to populate the
        properties of this descriptor (e.g., ``'ptb'`` for Psychtoolbox).

    Examples
    --------
    Get a list of available devices::

        import psychopy.sound as sound
        recordingDevicesList = sound.Microphone.getDevices()

    Get the low and high input latency of the first recording device::

        recordingDevice = recordingDevicesList[0]  # assume not empty
        inputLatencyLow, inputLatencyHigh = recordingDevice.inputLatency

    Get the device name as it may appear in the system control panel or sound
    settings::

        deviceName = recordingDevice.deviceName

    Specifying the device to use for capturing audio from a microphone::

        # get the first suitable capture device found by the sound engine
        recordingDevicesList = sound.Microphone.getDevices()
        recordingDevice = recordingDevicesList[0]

        # pass the descriptor to microphone to configure it
        mic = sound.Microphone(device=recordingDevice)
        mic.start()  # start recording sound

    """
    __slots__ = [
        '_deviceIndex',
        '_deviceName',
        '_hostAPIName',
        '_outputChannels',
        '_inputChannels',
        '_lowInputLatency',
        '_highInputLatency',
        '_lowOutputLatency',
        '_highOutputLatency',
        '_defaultSampleRate',
        '_audioLib'
    ]

    def __init__(self,
                 deviceIndex=-1,
                 deviceName=u'Null Device',
                 hostAPIName=u'Null Audio Driver',
                 outputChannels=0,
                 outputLatency=(0., 0.),
                 inputChannels=0,
                 inputLatency=(0., 0.),
                 defaultSampleRate=SAMPLE_RATE_48kHz,
                 audioLib=u''):

        # values based off Psychtoolbox audio device descriptors
        self.deviceIndex = deviceIndex
        self.deviceName = deviceName
        self.hostAPIName = hostAPIName
        self.outputChannels = outputChannels
        self.inputChannels = inputChannels
        self.inputLatency = inputLatency
        self.outputLatency = outputLatency
        self.defaultSampleRate = defaultSampleRate
        self.audioLib = audioLib

    def __repr__(self):
        return (f"AudioDeviceInfo(deviceIndex={self.deviceIndex}, "
                f"deviceName={self.deviceName}, "
                f"hostAPIName={self.hostAPIName}, "
                f"outputChannels={self.outputChannels}, "
                f"outputLatency={repr(self.outputLatency)}, "
                f"inputChannels={self.inputChannels}, "
                f"inputLatency={repr(self.inputLatency)}, "
                f"defaultSampleRate={self.defaultSampleRate}, "
                f"audioLib={repr(self.audioLib)})")

    @staticmethod
    def createFromPTBDesc(desc):
        """Create an `AudioDevice` instance with values populated using a
        descriptor (`dict`) returned from the PTB `audio.get_devices` API call.

        Parameters
        ----------
        desc : dict
            Audio device descriptor returned from Psychtoolbox's `get_devices`
            function.

        Returns
        -------
        AudioDeviceInfo
            Audio device descriptor with properties set using `desc`.

        """
        assert isinstance(desc, dict)

        # required fields, sanity check to see if something changed in PTB land
        reqFields = [
            'DeviceIndex', 'DeviceName', 'HostAudioAPIName', 'NrOutputChannels',
            'NrInputChannels', 'LowOutputLatency', 'HighOutputLatency',
            'LowInputLatency', 'HighInputLatency', 'DefaultSampleRate'
        ]

        assert all([field in desc.keys() for field in reqFields])

        audioDevDesc = AudioDeviceInfo(
            deviceIndex=desc['DeviceIndex'],
            deviceName=desc['DeviceName'],
            hostAPIName=desc['HostAudioAPIName'],
            outputChannels=desc['NrOutputChannels'],
            outputLatency=(desc['LowOutputLatency'], desc['HighOutputLatency']),
            inputChannels=desc['NrInputChannels'],
            inputLatency=(desc['LowInputLatency'], desc['HighInputLatency']),
            defaultSampleRate=desc['DefaultSampleRate'],
            audioLib=AUDIO_LIBRARY_PTB)  # queried with psychtoolbox

        return audioDevDesc

    @property
    def audioLib(self):
        """Audio library used to query device information (`str`)."""
        return self._audioLib

    @audioLib.setter
    def audioLib(self, value):
        self._audioLib = str(value)

    @property
    def inputChannels(self):
        """Number of input channels (`int`). If >0, this is likely a audio
        capture device.
        """
        return self._inputChannels

    @inputChannels.setter
    def inputChannels(self, value):
        self._inputChannels = int(value)

    @property
    def outputChannels(self):
        """Number of output channels (`int`). If >0, this is likely a audio
        playback device.
        """
        return self._outputChannels

    @outputChannels.setter
    def outputChannels(self, value):
        self._outputChannels = int(value)

    @property
    def deviceIndex(self):
        """Enumerated index (`int`) of the audio device."""
        return self._deviceIndex

    @deviceIndex.setter
    def deviceIndex(self, value):
        self._deviceIndex = int(value)

    @property
    def deviceName(self):
        """Human-readable name (`str`) for the audio device reported by the
        driver.
        """
        return self._deviceName

    @deviceName.setter
    def deviceName(self, value):
        self._deviceName = str(value)

    @property
    def hostAPIName(self):
        """Human-readable name (`str`) for the host API."""
        return self._hostAPIName

    @hostAPIName.setter
    def hostAPIName(self, value):
        self._hostAPIName = str(value)

    @property
    def outputLatency(self):
        """Low and high output latency in milliseconds `(low, high)`."""
        return self._lowOutputLatency, self._highOutputLatency

    @outputLatency.setter
    def outputLatency(self, value):
        assert len(value) == 2
        self._lowOutputLatency = float(value[0])
        self._highOutputLatency = float(value[1])

    @property
    def inputLatency(self):
        """Low and high input latency in milliseconds `(low, high)`."""
        return self._lowInputLatency, self._highInputLatency

    @inputLatency.setter
    def inputLatency(self, value):
        assert len(value) == 2
        self._lowInputLatency = float(value[0])
        self._highInputLatency = float(value[1])

    @property
    def defaultSampleRate(self):
        """Default sample rate in Hertz (Hz) for this device (`int`)."""
        return self._defaultSampleRate

    @defaultSampleRate.setter
    def defaultSampleRate(self, value):
        self._defaultSampleRate = int(value)

    @property
    def isPlayback(self):
        """`True` if this device is suitable for playback (`bool`)."""
        return self._outputChannels > 0

    @property
    def isCapture(self):
        """`True` if this device is suitable for capture (`bool`)."""
        return self._inputChannels > 0

    @property
    def isDuplex(self):
        """`True` if this device is suitable for capture and playback (`bool`).

        """
        return self.isPlayback and self.isCapture


class AudioDeviceStatus:
    """Descriptor for audio device status information.

    Properties of this class are standardized on the status information returned
    by Psychtoolbox. Other audio backends should try to populate these fields as
    best they can with their equivalent status values.

    Users should never instance this class themselves unless they have good
    reason to.

    Parameters
    ----------
    active : bool
        `True` if playback or recording has started, else `False`.
    state : int
        State of the device, either `1` for playback, `2` for recording or `3`
        for duplex (recording and playback).
    requestedStartTime : float
        Requested start time of the audio stream after the start of playback or
        recording.
    startTime : float
        The actual (real) start time of audio playback or recording.
    captureStartTime : float
        Estimate of the start time of audio capture. Only valid if audio capture
        is active. Usually, this time corresponds to the time when the first
        sound was captured.
    requestedStopTime : float
        Stop time requested when starting the stream.
    estimatedStopTime : float
        Estimated stop time given `requestedStopTime`.
    currentStreamTime : float
        Estimate of the time it will take for the most recently submitted sample
        to reach the speaker. Value is in absolute system time and reported for
        playback only.
    elapsedOutSamples : int
        Total number of samples submitted since the start of playback.
    positionSecs : float
        Current stream playback position in seconds this loop. Does not account
        for hardware of driver latency.
    recordedSecs : float
        Total amount of recorded sound data (in seconds) since start of capture.
    readSecs : float
        Total amount of sound data in seconds that has been fetched from the
        internal buffer.
    schedulePosition : float
        Current position in a running schedule in seconds.
    xRuns : int
        Number of dropouts due to buffer over- and under-runs. Such conditions
        can result is glitches during playback/recording. Even if the number
        remains zero, that does not mean that glitches did not occur.
    totalCalls : int
        **Debug** - Used for debugging the audio engine.
    timeFailed : float
        **Debug** - Used for debugging the audio engine.
    bufferSize : int
        **Debug** - Size of the buffer allocated to contain stream samples. Used
        for debugging the audio engine.
    cpuLoad : float
        Amount of load on the CPU imparted by the sound engine. Ranges between
        0.0 and 1.0 where 1.0 indicates maximum load on the core running the
        sound engine process.
    predictedLatency : float
        Latency for the given hardware and driver. This indicates how far ahead
        you need to start the device to ensure is starts at a scheduled time.
    latencyBias : float
        Additional latency bias added by the user.
    sampleRate : int
        Sample rate in Hertz (Hz) the playback recording is using.
    outDeviceIndex : int
        Enumerated index of the output device.
    inDeviceIndex : int
        Enumerated index of the input device.
    audioLib : str
        Identifier for the audio library which created this status.

    """
    __slots__ = [
        '_active',
        '_state',
        '_requestedStartTime',
        '_startTime',
        '_captureStartTime',
        '_requestedStopTime',
        '_estimatedStopTime',
        '_currentStreamTime',
        '_elapsedOutSamples',
        '_positionSecs',
        '_recordedSecs',
        '_readSecs',
        '_schedulePosition',
        '_xRuns',
        '_totalCalls',
        '_timeFailed',
        '_bufferSize',
        '_cpuLoad',
        '_predictedLatency',
        '_latencyBias',
        '_sampleRate',
        '_outDeviceIndex',
        '_inDeviceIndex',
        '_audioLib'
    ]

    def __init__(self,
                 active=0,
                 state=0,
                 requestedStartTime=0.0,
                 startTime=0.0,
                 captureStartTime=0.0,
                 requestedStopTime=0.0,
                 estimatedStopTime=0.0,
                 currentStreamTime=0.0,
                 elapsedOutSamples=0,
                 positionSecs=0.0,
                 recordedSecs=0.0,
                 readSecs=0.0,
                 schedulePosition=0.0,
                 xRuns=0,
                 totalCalls=0,
                 timeFailed=0,
                 bufferSize=0,
                 cpuLoad=0.0,
                 predictedLatency=0.0,
                 latencyBias=0.0,
                 sampleRate=SAMPLE_RATE_48kHz,
                 outDeviceIndex=0,
                 inDeviceIndex=0,
                 audioLib=u'Null Audio Library'):

        self.active = active
        self.state = state
        self.requestedStartTime = requestedStartTime
        self.startTime = startTime
        self.captureStartTime = captureStartTime
        self.requestedStopTime = requestedStopTime
        self.estimatedStopTime = estimatedStopTime
        self.currentStreamTime = currentStreamTime
        self.elapsedOutSamples = elapsedOutSamples
        self.positionSecs = positionSecs
        self.recordedSecs = recordedSecs
        self.readSecs = readSecs
        self.schedulePosition = schedulePosition
        self.xRuns = xRuns
        self.totalCalls = totalCalls
        self.timeFailed = timeFailed
        self.bufferSize = bufferSize
        self.cpuLoad = cpuLoad
        self.predictedLatency = predictedLatency
        self.latencyBias = latencyBias
        self.sampleRate = sampleRate
        self.outDeviceIndex = outDeviceIndex
        self.inDeviceIndex = inDeviceIndex
        self.audioLib = audioLib

    def __repr__(self):
        return (f"AudioDeviceStatus(active={self.active}, "
                f"state={self.state}, "
                f"requestedStartTime={self.requestedStartTime}, "
                f"startTime={self.startTime}, "
                f"captureStartTime={self.captureStartTime}, "
                f"requestedStopTime={self.requestedStopTime}, "
                f"estimatedStopTime={self.estimatedStopTime}, "
                f"currentStreamTime={self.currentStreamTime}, "
                f"elapsedOutSamples={self.elapsedOutSamples}, "
                f"positionSecs={self.positionSecs}, "
                f"recordedSecs={self.recordedSecs}, "
                f"readSecs={self.readSecs}, "
                f"schedulePosition={self.schedulePosition}, "
                f"xRuns={self.xRuns}, "
                f"totalCalls={self.totalCalls}, "
                f"timeFailed={self.timeFailed}, "
                f"bufferSize={self.bufferSize}, "
                f"cpuLoad={self.cpuLoad}, "
                f"predictedLatency={self.predictedLatency}, "
                f"latencyBias={self.latencyBias}, "
                f"sampleRate={self.sampleRate}, "
                f"outDeviceIndex={self.outDeviceIndex}, "
                f"inDeviceIndex={self.inDeviceIndex}, "
                f"audioLib={repr(self.audioLib)})")

    @staticmethod
    def createFromPTBDesc(desc):
        """Create an `AudioDeviceStatus` instance using a status descriptor
        returned by Psychtoolbox.

        Parameters
        ----------
        desc : dict
            Audio device status descriptor.

        Returns
        -------
        AudioDeviceStatus
            Audio device descriptor with properties set using `desc`.

        """
        audioStatusDesc = AudioDeviceStatus(
            active=desc['Active'],
            state=desc['State'],
            requestedStartTime=desc['RequestedStartTime'],
            startTime=desc['StartTime'],
            captureStartTime=desc['CaptureStartTime'],
            requestedStopTime=desc['RequestedStopTime'],
            estimatedStopTime=desc['EstimatedStopTime'],
            currentStreamTime=desc['CurrentStreamTime'],
            elapsedOutSamples=desc['ElapsedOutSamples'],
            positionSecs=desc['PositionSecs'],
            recordedSecs=desc['RecordedSecs'],
            readSecs=desc['ReadSecs'],
            schedulePosition=desc['SchedulePosition'],
            xRuns=desc['XRuns'],
            totalCalls=desc['TotalCalls'],
            timeFailed=desc['TimeFailed'],
            bufferSize=desc['BufferSize'],
            cpuLoad=desc['CPULoad'],
            predictedLatency=desc['PredictedLatency'],
            latencyBias=desc['LatencyBias'],
            sampleRate=desc['SampleRate'],
            outDeviceIndex=desc['OutDeviceIndex'],
            inDeviceIndex=desc['InDeviceIndex'],
            audioLib='ptb')

        return audioStatusDesc

    @property
    def audioLib(self):
        """Identifier for the audio library which created this status (`str`).
        """
        return self._audioLib

    @audioLib.setter
    def audioLib(self, value):
        self._audioLib = str(value)

    @property
    def active(self):
        """`True` if playback or recording has started (`bool`).

        """
        return self._active

    @active.setter
    def active(self, value):
        self._active = bool(value)

    @property
    def state(self):
        """State of the device (`int`). Either `1` for playback, `2` for
        recording or `3` for duplex (recording and playback).

        """
        return self._state

    @state.setter
    def state(self, value):
        self._state = int(value)

    @property
    def isPlayback(self):
        """`True` if this device is operating in playback mode (`bool`)."""
        return self._state == 1 or self._state == 3

    @property
    def isCapture(self):
        """`True` if this device is operating in capture mode (`bool`)."""
        return self._state == 2 or self._state == 3

    @property
    def isDuplex(self):
        """`True` if this device is operating capture and recording mode
        (`bool`).

        """
        return self._state == 3

    @property
    def requestedStartTime(self):
        """Requested start time of the audio stream after the start of playback
        or recording (`float`).

        """
        return self._requestedStartTime

    @requestedStartTime.setter
    def requestedStartTime(self, value):
        self._requestedStartTime = float(value)

    @property
    def startTime(self):
        """The actual (real) start time of audio playback or recording
        (`float`).

        """
        return self._startTime

    @startTime.setter
    def startTime(self, value):
        self._startTime = float(value)

    @property
    def captureStartTime(self):
        """Estimate of the start time of audio capture (`float`). Only valid if
        audio capture is active. Usually, this time corresponds to the time when
        the first sound was captured.

        """
        return self._startTime

    @captureStartTime.setter
    def captureStartTime(self, value):
        self._captureStartTime = float(value)

    @property
    def requestedStopTime(self):
        """Stop time requested when starting the stream (`float`)."""
        return self._requestedStopTime

    @requestedStopTime.setter
    def requestedStopTime(self, value):
        self._requestedStopTime = float(value)

    @property
    def estimatedStopTime(self):
        """Estimated stop time given `requestedStopTime` (`float`)."""
        return self._requestedStopTime

    @estimatedStopTime.setter
    def estimatedStopTime(self, value):
        self._estimatedStopTime = float(value)

    @property
    def currentStreamTime(self):
        """Estimate of the time it will take for the most recently submitted
        sample to reach the speaker (`float`). Value is in absolute system time
        and reported for playback mode only.

        """
        return self._currentStreamTime

    @currentStreamTime.setter
    def currentStreamTime(self, value):
        self._currentStreamTime = float(value)

    @property
    def elapsedOutSamples(self):
        """Total number of samples submitted since the start of playback
        (`int`).

        """
        return self._elapsedOutSamples

    @elapsedOutSamples.setter
    def elapsedOutSamples(self, value):
        self._elapsedOutSamples = int(value)

    @property
    def positionSecs(self):
        """Current stream playback position in seconds this loop (`float`). Does
        not account for hardware of driver latency.

        """
        return self._positionSecs

    @positionSecs.setter
    def positionSecs(self, value):
        self._positionSecs = float(value)

    @property
    def recordedSecs(self):
        """Total amount of recorded sound data (in seconds) since start of
        capture (`float`).

        """
        return self._recordedSecs

    @recordedSecs.setter
    def recordedSecs(self, value):
        self._recordedSecs = float(value)

    @property
    def readSecs(self):
        """Total amount of sound data in seconds that has been fetched from the
        internal buffer (`float`).

        """
        return self._readSecs

    @readSecs.setter
    def readSecs(self, value):
        self._readSecs = float(value)

    @property
    def schedulePosition(self):
        """Current position in a running schedule in seconds (`float`)."""
        return self._schedulePosition

    @schedulePosition.setter
    def schedulePosition(self, value):
        self._schedulePosition = float(value)

    @property
    def xRuns(self):
        """Number of dropouts due to buffer over- and under-runs (`int`). Such
        conditions can result is glitches during playback/recording. Even if the
        number remains zero, that does not mean that glitches did not occur.

        """
        return self._xRuns

    @xRuns.setter
    def xRuns(self, value):
        self._xRuns = int(value)

    @property
    def totalCalls(self):
        """**Debug** - Used for debugging the audio engine (`int`)."""
        return self._xRuns

    @totalCalls.setter
    def totalCalls(self, value):
        self._xRuns = int(value)

    @property
    def timeFailed(self):
        """**Debug** - Used for debugging the audio engine (`float`)."""
        return self._timeFailed

    @timeFailed.setter
    def timeFailed(self, value):
        self._timeFailed = float(value)

    @property
    def bufferSize(self):
        """**Debug** - Size of the buffer allocated to contain stream samples.
        Used for debugging the audio engine.

        """
        return self._bufferSize

    @bufferSize.setter
    def bufferSize(self, value):
        self._bufferSize = int(value)

    @property
    def cpuLoad(self):
        """Amount of load on the CPU imparted by the sound engine (`float`).
        Ranges between 0.0 and 1.0 where 1.0 indicates maximum load on the core
        running the sound engine process.

        """
        return self._cpuLoad

    @cpuLoad.setter
    def cpuLoad(self, value):
        self._cpuLoad = float(value)

    @property
    def predictedLatency(self):
        """Latency for the given hardware and driver (`float`). This indicates
        how far ahead you need to start the device to ensure is starts at a
        scheduled time.

        """
        return self._predictedLatency

    @predictedLatency.setter
    def predictedLatency(self, value):
        self._predictedLatency = float(value)

    @property
    def latencyBias(self):
        """Additional latency bias added by the user (`float`)."""
        return self._latencyBias

    @latencyBias.setter
    def latencyBias(self, value):
        self._latencyBias = float(value)

    @property
    def sampleRate(self):
        """Sample rate in Hertz (Hz) the playback recording is using (`int`)."""
        return self._sampleRate

    @sampleRate.setter
    def sampleRate(self, value):
        self._sampleRate = int(value)

    @property
    def outDeviceIndex(self):
        """Enumerated index of the output device (`int`)."""
        return self._outDeviceIndex

    @outDeviceIndex.setter
    def outDeviceIndex(self, value):
        self._outDeviceIndex = int(value)

    @property
    def inDeviceIndex(self):
        """Enumerated index of the input device (`int`)."""
        return self._inDeviceIndex

    @inDeviceIndex.setter
    def inDeviceIndex(self, value):
        self._inDeviceIndex = int(value)


# These are used as sentinels or for testing. Instancing these here behaves as
# a self-test, providing coverage to most of the setter methods when this module
# is imported.
#
NULL_AUDIO_DEVICE = AudioDeviceInfo()
NULL_AUDIO_DEVICE_STATUS = AudioDeviceStatus()

if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
There are many ways to generate counter-phase, e.g. vary the contrast of
a grating sinusoidally between 1 and -1, take 2 gratings in opposite phase
overlaid and vary the opacity of the upper one between 1: 0, or take two
gratings overlaid with the upper one of 0.5 opacity and drift them
in opposite directions.

This script takes the first approach as a test of how fast
contrast textures are being rewritten to the graphics card
"""
from psychopy import core, visual, event
from numpy import sin, pi

# Create a window to draw in
win = visual.Window((600, 600), allowGUI=False, monitor='testMonitor', units='deg')

# Initialize some stimuli
grating1 = visual.GratingStim(
    win, tex="sin", mask="circle", texRes=128,
    color='white', size=5, sf=2, ori=45, depth=0.5, autoLog=False)
message = visual.TextStim(
    win, text='Any key to quit',
    pos=(-0.95, -0.95), units='norm',
    anchorVert='bottom', anchorHoriz='left')

trialClock = core.Clock()
t = 0
while not event.getKeys() and t < 20:  # quits after 20 secs
    t = trialClock.getTime()

    grating1.contrast = sin(t * pi * 2)
    grating1.draw()
    message.draw()

    win.flip()

win.close()
core.quit()

# The contents of this file are in the public domain.



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Functions and classes related to variable type conversion
"""
import numpy


def float_uint8(inarray):
    """Converts arrays, lists, tuples and floats ranging -1:1
    into an array of Uint8s ranging 0:255

    >>> float_uint8(-1)
    0
    >>> float_uint8(0)
    128

    """
    retVal = numpy.around(255 * (0.5 + 0.5 * numpy.asarray(inarray)))
    return retVal.astype(numpy.uint8)


def float_uint16(inarray):
    """Converts arrays, lists, tuples and floats ranging -1:1
    into an array of Uint16s ranging 0:2^16

    >>> float_uint16(-1)
    0
    >>> float_uint16(0)
    32768

    """
    i16max = 2**16 - 1
    retVal = numpy.around(i16max * (1.0 + numpy.asarray(inarray)) / 2.0)
    return retVal.astype(numpy.uint16)


def uint8_float(inarray):
    """Converts arrays, lists, tuples and UINTs ranging 0:255
    into an array of floats ranging -1:1

    >>> uint8_float(0)
    -1.0
    >>> uint8_float(128)
    0.0

    """
    return numpy.asarray(inarray, 'f')/127.5 - 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Part of the PsychoPy library
Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
Distributed under the terms of the GNU General Public License (GPL).
"""

from pathlib import Path
from xml.etree.ElementTree import Element

from psychopy import prefs
from psychopy.constants import FOREVER
from ..params import Param
from psychopy.experiment.utils import CodeGenerationException
from psychopy.experiment.utils import unescapedDollarSign_re
from psychopy.experiment.params import getCodeFromParamStr
from psychopy.alerts import alerttools
from psychopy.colors import nonAlphaSpaces

from psychopy.localization import _translate, _localized


class BaseComponent:
    """A template for components, defining the methods to be overridden"""
    # override the categories property below
    # an attribute of the class, determines the section in the components panel

    categories = ['Custom']
    targets = []
    iconFile = Path(__file__).parent / "unknown" / "unknown.png"
    tooltip = ""

    def __init__(self, exp, parentName, name='',
                 startType='time (s)', startVal='',
                 stopType='duration (s)', stopVal='',
                 startEstim='', durationEstim='',
                 saveStartStop=True, syncScreenRefresh=False,
                 disabled=False):
        self.type = 'Base'
        self.exp = exp  # so we can access the experiment if necess
        self.parentName = parentName  # to access the routine too if needed

        self.params = {}
        self.depends = []  # allows params to turn each other off/on
        """{
         "dependsOn": "shape",
         "condition": "=='n vertices",
         "param": "n vertices",
         "true": "enable",  # what to do with param if condition is True
         "false": "disable",  # permitted: hide, show, enable, disable
         }"""
        self.order = ['name', 'startVal', 'startEstim', 'startType', 'stopVal', 'durationEstim', 'stopType']  # name first, then timing, then others

        msg = _translate(
            "Name of this component (alphanumeric or _, no spaces)")
        self.params['name'] = Param(name,
            valType='code', inputType="single", categ='Basic',
            hint=msg,
            label=_localized['name'])

        msg = _translate("How do you want to define your start point?")
        self.params['startType'] = Param(startType,
            valType='str', inputType="choice", categ='Basic',
            allowedVals=['time (s)', 'frame N', 'condition'],
            hint=msg, direct=False,
            label=_localized['startType'])

        msg = _translate("How do you want to define your end point?")
        self.params['stopType'] = Param(stopType,
            valType='str', inputType="choice", categ='Basic',
            allowedVals=['duration (s)', 'duration (frames)', 'time (s)',
                         'frame N', 'condition'],
            hint=msg, direct=False,
            label=_localized['stopType'])

        self.params['startVal'] = Param(startVal,
            valType='code', inputType="single", categ='Basic',
            hint=_translate("When does the component start?"), allowedTypes=[],
            label=_localized['startVal'])

        self.params['stopVal'] = Param(stopVal,
            valType='code', inputType="single", categ='Basic',
            updates='constant', allowedUpdates=[], allowedTypes=[],
            hint=_translate("When does the component end? (blank is endless)"),
            label=_localized['stopVal'])

        msg = _translate("(Optional) expected start (s), purely for "
                         "representing in the timeline")
        self.params['startEstim'] = Param(startEstim,
            valType='code', inputType="single", categ='Basic',
            hint=msg, allowedTypes=[], direct=False,
            label=_localized['startEstim'])

        msg = _translate("(Optional) expected duration (s), purely for "
                         "representing in the timeline")
        self.params['durationEstim'] = Param(durationEstim,
            valType='code', inputType="single", categ='Basic',
            hint=msg, allowedTypes=[], direct=False,
            label=_localized['durationEstim'])

        msg = _translate("Store the onset/offset times in the data file "
                         "(as well as in the log file).")
        self.params['saveStartStop'] = Param(saveStartStop,
            valType='bool', inputType="bool", categ='Data',
            hint=msg, allowedTypes=[],
            label=_translate('Save onset/offset times'))

        msg = _translate("Synchronize times with screen refresh (good for "
                         "visual stimuli and responses based on them)")
        self.params['syncScreenRefresh'] = Param(syncScreenRefresh,
            valType='bool', inputType="bool", categ="Data",
            hint=msg, allowedTypes=[],
            label=_translate('Sync timing with screen refresh'))

        msg = _translate("Disable this component")
        self.params['disabled'] = Param(disabled,
            valType='bool', inputType="bool", categ="Testing",
            hint=msg, allowedTypes=[], direct=False,
            label=_translate('Disable component'))

    @property
    def _xml(self):
        # Make root element
        element = Element(self.__class__.__name__)
        element.set("name", self.params['name'].val)
        # Add an element for each parameter
        for key, param in sorted(self.params.items()):
            # Create node
            paramNode = param._xml
            paramNode.set("name", key)
            # Add node
            element.append(paramNode)

        return element

    def __repr__(self):
        _rep = "psychopy.experiment.components.%s(name='%s', exp=%s)"
        return _rep % (self.__class__.__name__, self.name, self.exp)

    def integrityCheck(self):
        """
        Run component integrity checks for non-visual components
        """
        alerttools.testDisabled(self)
        alerttools.testStartEndTiming(self)

    def _dubiousConstantUpdates(self):
        """Return a list of fields in component that are set to be constant
        but seem intended to be dynamic. Some code fields are constant, and
        some denoted as code by $ are constant.
        """
        warnings = []
        # treat expInfo as likely to be constant; also treat its keys as
        # constant because its handy to make a short-cut in code:
        # exec(key+'=expInfo[key]')
        expInfo = self.exp.settings.getInfo()
        keywords = self.exp.namespace.nonUserBuilder[:]
        keywords.extend(['expInfo'] + list(expInfo.keys()))
        reserved = set(keywords).difference({'random', 'rand'})
        for key in self.params:
            field = self.params[key]
            if (not hasattr(field, 'val') or
                    not isinstance(field.val, str)):
                continue  # continue == no problem, no warning
            if not (field.allowedUpdates and
                    isinstance(field.allowedUpdates, list) and
                    len(field.allowedUpdates) and
                    field.updates == 'constant'):
                continue
            # now have only non-empty, possibly-code, and 'constant' updating
            if field.valType == 'str':
                if not bool(unescapedDollarSign_re.search(field.val)):
                    continue
                code = getCodeFromParamStr(field.val)
            elif field.valType == 'code':
                code = field.val
            else:
                continue
            # get var names in the code; no names == constant
            try:
                names = compile(code, '', 'eval').co_names
            except SyntaxError:
                continue
            # ignore reserved words:
            if not set(names).difference(reserved):
                continue
            warnings.append((field, key))
        return warnings or [(None, None)]

    def writeInitCode(self, buff):
        """Write any code that a component needs that should only ever be done
        at start of an experiment, BEFORE window creation.
        """
        pass

    def writeStartCode(self, buff):
        """Write any code that a component needs that should only ever be done
        at start of an experiment, AFTER window creation.
        """
        # e.g., create a data subdirectory unique to that component type.
        # Note: settings.writeStartCode() is done first, then
        # Routine.writeStartCode() will call this method for each component in
        # each routine
        pass

    def writeFrameCode(self, buff):
        """Write the code that will be called every frame
        """
        pass

    def writeRoutineStartCode(self, buff):
        """Write the code that will be called at the beginning of
        a routine (e.g. to update stimulus parameters)
        """
        self.writeParamUpdates(buff, 'set every repeat')

    def writeRoutineStartCodeJS(self, buff):
        """Same as writeRoutineStartCode, but for JS
        """
        self.writeParamUpdatesJS(buff, 'set every repeat')

    def writeRoutineEndCode(self, buff):
        """Write the code that will be called at the end of
        a routine (e.g. to save data)
        """
        if 'saveStartStop' in self.params and self.params['saveStartStop'].val:

            # what loop are we in (or thisExp)?
            if len(self.exp.flow._loopList):
                currLoop = self.exp.flow._loopList[-1]  # last (outer-most) loop
            else:
                currLoop = self.exp._expHandler

            if 'Stair' in currLoop.type:
                addDataFunc = 'addOtherData'
            else:
                addDataFunc = 'addData'

            loop = currLoop.params['name']
            name = self.params['name']
            if self.params['syncScreenRefresh'].val:
                code = (
                    f"{loop}.{addDataFunc}('{name}.started', {name}.tStartRefresh)\n"
                    f"{loop}.{addDataFunc}('{name}.stopped', {name}.tStopRefresh)\n"
                )
            else:
                code = (
                    f"{loop}.{addDataFunc}('{name}.started', {name}.tStart)\n"
                    f"{loop}.{addDataFunc}('{name}.stopped', {name}.tStop)\n"
                )

            buff.writeIndentedLines(code)

    def writeRoutineEndCodeJS(self, buff):
        """Write the code that will be called at the end of
        a routine (e.g. to save data)
        """
        pass

    def writeExperimentEndCode(self, buff):
        """Write the code that will be called at the end of
        an experiment (e.g. save log files or reset hardware)
        """
        pass

    def writeStartTestCode(self, buff):
        """Test whether we need to start
        """
        if self.params['syncScreenRefresh']:
            tCompare = 'tThisFlip'
        else:
            tCompare = 't'
        params = self.params
        t = tCompare
        if self.params['startType'].val == 'time (s)':
            # if startVal is an empty string then set to be 0.0
            if (isinstance(self.params['startVal'].val, str) and
                    not self.params['startVal'].val.strip()):
                self.params['startVal'].val = '0.0'
            code = (f"if {params['name']}.status == NOT_STARTED and "
                    f"{t} >= {params['startVal']}-frameTolerance:\n")
        elif self.params['startType'].val == 'frame N':
            code = (f"if {params['name']}.status == NOT_STARTED and "
                    f"frameN >= {params['startVal']}:\n")
        elif self.params['startType'].val == 'condition':
            code = (f"if {params['name']}.status == NOT_STARTED and "
                    f"{params['startVal']}:\n")
        else:
            msg = f"Not a known startType ({params['startVal']}) for {params['name']}"
            raise CodeGenerationException(msg % self.params)

        buff.writeIndented(code)

        buff.setIndentLevel(+1, relative=True)
        params = self.params
        code = (f"# keep track of start time/frame for later\n"
                f"{params['name']}.frameNStart = frameN  # exact frame index\n"
                f"{params['name']}.tStart = t  # local t and not account for scr refresh\n"
                f"{params['name']}.tStartRefresh = tThisFlipGlobal  # on global time\n")
        if self.type != "Sound":  # for sounds, don't update to actual frame time
                code += (f"win.timeOnFlip({params['name']}, 'tStartRefresh')"
                         f"  # time at next scr refresh\n")
        buff.writeIndentedLines(code)

    def writeStartTestCodeJS(self, buff):
        """Test whether we need to start
        """
        params = self.params
        if self.params['startType'].val == 'time (s)':
            # if startVal is an empty string then set to be 0.0
            if (isinstance(self.params['startVal'].val, str) and
                    not self.params['startVal'].val.strip()):
                self.params['startVal'].val = '0.0'
            code = (f"if (t >= {params['startVal']} "
                    f"&& {params['name']}.status === PsychoJS.Status.NOT_STARTED) {{\n")
        elif self.params['startType'].val == 'frame N':
            code = (f"if (frameN >= {params['startVal']} "
                    f"&& {params['name']}.status === PsychoJS.Status.NOT_STARTED) {{\n")
        elif self.params['startType'].val == 'condition':
            code = (f"if (({params['startVal']}) "
                    f"&& {params['name']}.status === PsychoJS.Status.NOT_STARTED) {{\n")
        else:
            msg = f"Not a known startType ({params['startVal']}) for {params['name']}"
            raise CodeGenerationException(msg)

        buff.writeIndented(code)

        buff.setIndentLevel(+1, relative=True)
        code = (f"// keep track of start time/frame for later\n"
                f"{params['name']}.tStart = t;  // (not accounting for frame time here)\n"
                f"{params['name']}.frameNStart = frameN;  // exact frame index\n\n")
        buff.writeIndentedLines(code)

    def writeStopTestCode(self, buff):
        """Test whether we need to stop
        """
        params = self.params
        buff.writeIndentedLines(f"if {params['name']}.status == STARTED:\n")
        buff.setIndentLevel(+1, relative=True)

        # If start time is blank ad stop is a duration, raise alert
        if self.params['stopType'] in ('duration (s)', 'duration (frames)'):
            if ('startVal' not in self.params) or (self.params['startVal'] in ("", "None", None)):
                alerttools.alert(4120, strFields={'component': self.params['name']})

        if self.params['stopType'].val == 'time (s)':
            code = (f"# is it time to stop? (based on local clock)\n"
                    f"if tThisFlip > {params['stopVal']}-frameTolerance:\n"
                    )
        # duration in time (s)
        elif (self.params['stopType'].val == 'duration (s)'):
            code = (f"# is it time to stop? (based on global clock, using actual start)\n"
                    f"if tThisFlipGlobal > {params['name']}.tStartRefresh + {params['stopVal']}-frameTolerance:\n")
        elif self.params['stopType'].val == 'duration (frames)':
            code = (f"if frameN >= ({params['name']}.frameNStart + {params['stopVal']}):\n")
        elif self.params['stopType'].val == 'frame N':
            code = f"if frameN >= {params['stopVal']}:\n"
        elif self.params['stopType'].val == 'condition':
            code = f"if bool({params['stopVal']}):\n"
        else:
            msg = (f"Didn't write any stop line for startType={params['startType']}, "
                   f"stopType={params['stopType']}")
            raise CodeGenerationException(msg)

        buff.writeIndentedLines(code)
        buff.setIndentLevel(+1, relative=True)
        code = (f"# keep track of stop time/frame for later\n"
                f"{params['name']}.tStop = t  # not accounting for scr refresh\n"
                f"{params['name']}.frameNStop = frameN  # exact frame index\n"
                f"win.timeOnFlip({params['name']}, 'tStopRefresh')"
                f"  # time at next scr refresh\n")
        buff.writeIndentedLines(code)

    def writeStopTestCodeJS(self, buff):
        """Test whether we need to stop
        """
        params = self.params
        if self.params['stopType'].val == 'time (s)':
            code = (f"frameRemains = {params['stopVal']} "
                    f" - psychoJS.window.monitorFramePeriod * 0.75;"
                    f"  // most of one frame period left\n"
                    f"if (({params['name']}.status === PsychoJS.Status.STARTED || {params['name']}.status === PsychoJS.Status.FINISHED) "
                    f"&& t >= frameRemains) {{\n")
        # duration in time (s)
        elif (self.params['stopType'].val == 'duration (s)' and
              self.params['startType'].val == 'time (s)'):
            code = (f"frameRemains = {params['startVal']} + {params['stopVal']}"
                    f" - psychoJS.window.monitorFramePeriod * 0.75;"
                    f"  // most of one frame period left\n"
                    f"if ({params['name']}.status === PsychoJS.Status.STARTED "
                    f"&& t >= frameRemains) {{\n")
        # start at frame and end with duratio (need to use approximate)
        elif self.params['stopType'].val == 'duration (s)':
            code = (f"if ({params['name']}.status === PsychoJS.Status.STARTED "
                    f"&& t >= ({params['name']}.tStart + {params['stopVal']})) {{\n")
        elif self.params['stopType'].val == 'duration (frames)':
            code = (f"if ({params['name']}.status === PsychoJS.Status.STARTED "
                    f"&& frameN >= ({params['name']}.frameNStart + {params['stopVal']})) {{\n")
        elif self.params['stopType'].val == 'frame N':
            code = (f"if ({params['name']}.status === PsychoJS.Status.STARTED "
                    f"&& frameN >= {params['stopVal']}) {{\n")
        elif self.params['stopType'].val == 'condition':
            code = (f"if ({params['name']}.status === PsychoJS.Status.STARTED "
                    f"&& Boolean({params['stopVal']})) {{\n")
        else:
            msg = (f"Didn't write any stop line for startType="
                   f"{params['startType']}, "
                   f"stopType={params['stopType']}")
            raise CodeGenerationException(msg)

        buff.writeIndentedLines(code)
        buff.setIndentLevel(+1, relative=True)

    def writeParamUpdates(self, buff, updateType, paramNames=None,
                          target="PsychoPy"):
        """write updates to the buffer for each parameter that needs it
        updateType can be 'experiment', 'routine' or 'frame'
        """
        if paramNames is None:
            paramNames = list(self.params.keys())
        for thisParamName in paramNames:
            if thisParamName == 'advancedParams':
                continue  # advancedParams is not really a parameter itself
            thisParam = self.params[thisParamName]
            if thisParam.updates == updateType:
                self.writeParamUpdate(
                    buff, self.params['name'],
                    thisParamName, thisParam, thisParam.updates,
                    target=target)

    def writeParamUpdatesJS(self, buff, updateType, paramNames=None):
        """Pass this to the standard writeParamUpdates but with new 'target'
        """
        self.writeParamUpdates(buff, updateType, paramNames,
                               target="PsychoJS")

    def writeParamUpdate(self, buff, compName, paramName, val, updateType,
                         params=None, target="PsychoPy"):
        """Writes an update string for a single parameter.
        This should not need overriding for different components - try to keep
        constant
        """
        if params is None:
            params = self.params
        # first work out the name for the set____() function call
        if paramName == 'advancedParams':
            return  # advancedParams is not really a parameter itself
        elif paramName == 'letterHeight':
            paramCaps = 'Height'  # setHeight for TextStim
        elif paramName == 'image' and self.getType() == 'PatchComponent':
            paramCaps = 'Tex'  # setTex for PatchStim
        elif paramName == 'sf':
            paramCaps = 'SF'  # setSF, not SetSf
        elif paramName == 'coherence':
            paramCaps = 'FieldCoherence'
        elif paramName == 'fieldPos':
            paramCaps = 'FieldPos'
        else:
            paramCaps = paramName[0].capitalize() + paramName[1:]

        # code conversions for PsychoJS
        if target == 'PsychoJS':
            endStr = ';'
            try:
                valStr = str(val).strip()
            except TypeError:
                if isinstance(val, Param):
                    val = val.val
                raise TypeError(f"Value of parameter {paramName} of component {compName} "
                                f"could not be converted to JS. Value is {val}")
            # convert (0,0.5) to [0,0.5] but don't convert "rand()" to "rand[]"
            if valStr.startswith("(") and valStr.endswith(")"):
                valStr = valStr.replace("(", "[", 1)
                valStr = valStr[::-1].replace(")", "]", 1)[
                         ::-1]  # replace from right
            # filenames (e.g. for image) need to be loaded from resources
            if paramName in ["sound"]:
                valStr = (f"psychoJS.resourceManager.getResource({valStr})")
        else:
            endStr = ''

        # then write the line
        if updateType == 'set every frame' and target == 'PsychoPy':
            loggingStr = ', log=False'
        elif updateType == 'set every frame' and target == 'PsychoJS':
            loggingStr = ', false'  # don't give the keyword 'log' in JS
        else:
            loggingStr = ''

        if target == 'PsychoPy':
            if paramName == 'color':
                buff.writeIndented(f"{compName}.setColor({params['color']}, colorSpace={params['colorSpace']}")
                buff.write(f"{loggingStr}){endStr}\n")
            elif paramName == 'sound':
                stopVal = params['stopVal'].val
                if stopVal in ['', None, -1, 'None']:
                    stopVal = '-1'
                buff.writeIndented(f"{compName}.setSound({params['sound']}, secs={stopVal}){endStr}\n")
            elif paramName == 'movie':
                # we're going to do this for now ...
                if params['units'].val == 'from exp settings':
                    unitsStr = "units=''"
                else:
                    unitsStr = "units=%(units)s" % params

                if params['backend'].val == 'moviepy':
                    code = ("%s = visual.MovieStim3(\n" % params['name'] +
                            "    win=win, name='%s', %s,\n" % (
                                params['name'], unitsStr) +
                            "    noAudio = %(No audio)s,\n" % params)
                elif params['backend'].val == 'avbin':
                    code = ("%s = visual.MovieStim(\n" % params['name'] +
                            "    win=win, name='%s', %s,\n" % (
                                params['name'], unitsStr))
                elif params['backend'].val == 'vlc':
                    code = ("%s = visual.VlcMovieStim(\n" % params['name'] +
                            "    win=win, name='%s', %s,\n" % (
                                params['name'], unitsStr))
                else:
                    code = ("%s = visual.MovieStim2(\n" % params['name'] +
                            "    win=win, name='%s', %s,\n" % (
                                params['name'], unitsStr) +
                            "    noAudio=%(No audio)s,\n" % params)

                code += ("    filename=%(movie)s,\n"
                         "    ori=%(ori)s, pos=%(pos)s, opacity=%(opacity)s,\n"
                         "    loop=%(loop)s, anchor=%(anchor)s,\n"
                         % params)

                buff.writeIndentedLines(code)

                if params['size'].val != '':
                    buff.writeIndented("    size=%(size)s,\n" % params)

                depth = -self.getPosInRoutine()
                code = ("    depth=%.1f,\n"
                        "    )\n")
                buff.writeIndentedLines(code % depth)

            else:
                buff.writeIndented(f"{compName}.set{paramCaps}({val}{loggingStr}){endStr}\n")
        elif target == 'PsychoJS':
            # write the line
            if paramName == 'color':
                buff.writeIndented(f"{compName}.setColor(new util.Color({params['color']})")
                buff.write(f"{loggingStr}){endStr}\n")
            elif paramName == 'fillColor':
                buff.writeIndented(f"{compName}.setFillColor(new util.Color({params['fillColor']})")
                buff.write(f"{loggingStr}){endStr}\n")
            elif paramName == 'lineColor':
                buff.writeIndented(f"{compName}.setLineColor(new util.Color({params['lineColor']})")
                buff.write(f"{loggingStr}){endStr}\n")
            elif paramName == 'sound':
                stopVal = params['stopVal']
                if stopVal in ['', None, -1, 'None']:
                    stopVal = '-1'
                buff.writeIndented(f"{compName}.setSound({params['sound']}, secs={stopVal}){endStr}\n")
            elif compName == 'eeg_marker':
                # This allows the eeg_marker to be updated by a code component or a conditions file
                # There is no setMarker_label or setMarker_value function in the eeg_marker object
                # The marker label and value are set by the variables set in the dialogue
                pass
            else:
                buff.writeIndented(f"{compName}.set{paramCaps}({val}{loggingStr}){endStr}\n")

    def checkNeedToUpdate(self, updateType):
        """Determine whether this component has any parameters set to repeat
        at this level

        usage::
            True/False = checkNeedToUpdate(self, updateType)

        """
        for thisParamName in self.params:
            if thisParamName == 'advancedParams':
                continue
            thisParam = self.params[thisParamName]
            if thisParam.updates == updateType:
                return True

        return False

    def getStartAndDuration(self):
        """Determine the start and duration of the stimulus
        purely for Routine rendering purposes in the app (does not affect
        actual drawing during the experiment)

        start, duration, nonSlipSafe = component.getStartAndDuration()

        nonSlipSafe indicates that the component's duration is a known fixed
        value and can be used in non-slip global clock timing (e.g for fMRI)
        """
        if not 'startType' in self.params:
            # this component does not have any start
            return None, None, True

        # deduce a start time (s) if possible
        startType = self.params['startType'].val
        numericStart = canBeNumeric(self.params['startVal'].val)

        if canBeNumeric(self.params['startEstim'].val):
            startTime = float(self.params['startEstim'].val)
        elif startType == 'time (s)' and numericStart:
            startTime = float(self.params['startVal'].val)
        else:
            startTime = None

        if 'stopType' not in self.params:
            # this component does not have any stop
            return startTime, 0, numericStart

        # deduce stop time (s) if possible
        stopType = self.params['stopType'].val
        numericStop = canBeNumeric(self.params['stopVal'].val)

        if stopType == 'time (s)' and numericStop:
            duration = float(self.params['stopVal'].val) - (startTime or 0)
        elif stopType == 'duration (s)' and numericStop:
            duration = float(self.params['stopVal'].val)
        else:
            # deduce duration (s) if possible. Duration used because component
            # time icon needs width
            if canBeNumeric(self.params['durationEstim'].val):
                duration = float(self.params['durationEstim'].val)
            elif self.params['stopVal'].val in ['', '-1', 'None']:
                duration = FOREVER  # infinite duration
            else:
                duration = None

        nonSlipSafe = numericStop and (numericStart or stopType == 'time (s)')
        return startTime, duration, nonSlipSafe

    def getPosInRoutine(self):
        """Find the index (position) in the parent Routine (0 for top)
        """
        routine = self.exp.routines[self.parentName]
        return routine.index(self)

    def getType(self):
        """Returns the name of the current object class"""
        return self.__class__.__name__

    def getShortType(self):
        """Replaces word component with empty string"""
        return self.getType().replace('Component', '')

    @property
    def name(self):
        return self.params['name'].val

    @name.setter
    def name(self, value):
        self.params['name'].val = value

    @property
    def disabled(self):
        return bool(self.params['disabled'])

    @disabled.setter
    def disabled(self, value):
        self.params['disabled'].val = value


class BaseVisualComponent(BaseComponent):
    """Base class for most visual stimuli
    """

    categories = ['Stimuli']
    targets = []
    iconFile = Path(__file__).parent / "unknown" / "unknown.png"
    tooltip = ""

    def __init__(self, exp, parentName, name='',
                 units='from exp settings', color='white', fillColor="", borderColor="",
                 pos=(0, 0), size=(0, 0), ori=0, colorSpace='rgb', opacity="", contrast=1,
                 startType='time (s)', startVal='',
                 stopType='duration (s)', stopVal='',
                 startEstim='', durationEstim='',
                 saveStartStop=True, syncScreenRefresh=True):

        super(BaseVisualComponent, self).__init__(
            exp, parentName, name,
            startType=startType, startVal=startVal,
            stopType=stopType, stopVal=stopVal,
            startEstim=startEstim, durationEstim=durationEstim,
            saveStartStop=saveStartStop,
            syncScreenRefresh=syncScreenRefresh)

        self.exp.requirePsychopyLibs(
            ['visual'])  # needs this psychopy lib to operate

        self.order += [
            "color",
            "fillColor",
            "borderColor",
            "colorSpace",
            "opacity",
            "size",
            "pos",
            "units",
            "anchor",
            "ori",
        ]

        msg = _translate("Units of dimensions for this stimulus")
        self.params['units'] = Param(units,
            valType='str', inputType="choice", categ='Layout',
            allowedVals=['from exp settings', 'deg', 'cm', 'pix', 'norm',
                         'height', 'degFlatPos', 'degFlat'],
            hint=msg,
            label=_localized['units'])

        msg = _translate("Foreground color of this stimulus (e.g. $[1,1,0], red )")
        self.params['color'] = Param(color,
            valType='color', inputType="color", categ='Appearance',
            allowedTypes=[],
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['color'])

        msg = _translate("In what format (color space) have you specified "
                         "the colors? (rgb, dkl, lms, hsv)")
        self.params['colorSpace'] = Param(colorSpace,
            valType='str', inputType="choice", categ='Appearance',
            allowedVals=['rgb', 'dkl', 'lms', 'hsv'],
            updates='constant',
            hint=msg,
            label=_localized['colorSpace'])

        msg = _translate("Fill color of this stimulus (e.g. $[1,1,0], red )")
        self.params['fillColor'] = Param(fillColor,
            valType='color', inputType="color", categ='Appearance',
            updates='constant', allowedTypes=[],
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['fillColor'])

        msg = _translate("Color of this stimulus (e.g. $[1,1,0], red )")
        self.params['borderColor'] = Param(borderColor,
            valType='color', inputType="color", categ='Appearance',
            updates='constant',allowedTypes=[],
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['borderColor'])

        msg = _translate("Opacity of the stimulus (1=opaque, 0=fully transparent, 0.5=translucent). "
                         "Leave blank for each color to have its own opacity (recommended if any color is None).")
        self.params['opacity'] = Param(opacity,
            valType='num', inputType="single", categ='Appearance',
            updates='constant', allowedTypes=[],
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['opacity'])

        msg = _translate("Contrast of the stimulus (1.0=unchanged contrast, "
                         "0.5=decrease contrast, 0.0=uniform/no contrast, "
                         "-0.5=slightly inverted, -1.0=totally inverted)")
        self.params['contrast'] = Param(contrast,
            valType='num', inputType='single', allowedTypes=[], categ='Appearance',
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['contrast'])

        msg = _translate("Position of this stimulus (e.g. [1,2] )")
        self.params['pos'] = Param(pos,
            valType='list', inputType="single", categ='Layout',
            updates='constant', allowedTypes=[],
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['pos'])

        msg = _translate("Size of this stimulus (either a single value or "
                         "x,y pair, e.g. 2.5, [1,2] ")
        self.params['size'] = Param(size,
            valType='list', inputType="single", categ='Layout',
            updates='constant', allowedTypes=[],
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['size'])

        self.params['ori'] = Param(ori,
            valType='num', inputType="spin", categ='Layout',
            updates='constant', allowedTypes=[], allowedVals=[-360,360],
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=_translate("Orientation of this stimulus (in deg)"),
            label=_localized['ori'])

        self.params['syncScreenRefresh'].readOnly = True

    def integrityCheck(self):
        """
        Run component integrity checks.
        """
        super().integrityCheck()  # run parent class checks first

        win = alerttools.TestWin(self.exp)

        # get units for this stimulus
        if 'units' in self.params:  # e.g. BrushComponent doesn't have this
            units = self.params['units'].val
        else:
            units = None
        if units == 'use experiment settings':
            units = self.exp.settings.params[
                'Units'].val  # this 1 uppercase
        if not units or units == 'use preferences':
            units = prefs.general['units']

        # tests for visual stimuli
        alerttools.testSize(self, win, units)
        alerttools.testPos(self, win, units)
        alerttools.testAchievableVisualOnsetOffset(self)
        alerttools.testValidVisualStimTiming(self)
        alerttools.testFramesAsInt(self)

    def writeFrameCode(self, buff):
        """Write the code that will be called every frame
        """
        params = self.params
        buff.writeIndented(f"\n")
        buff.writeIndented(f"# *{params['name']}* updates\n")
        # writes an if statement to determine whether to draw etc
        self.writeStartTestCode(buff)
        buff.writeIndented(f"{params['name']}.setAutoDraw(True)\n")
        # to get out of the if statement
        buff.setIndentLevel(-1, relative=True)

        # test for stop (only if there was some setting for duration or stop)
        if self.params['stopVal'].val not in ('', None, -1, 'None'):
            # writes an if statement to determine whether to draw etc
            self.writeStopTestCode(buff)
            buff.writeIndented(f"{params['name']}.setAutoDraw(False)\n")
            # to get out of the if statement
            buff.setIndentLevel(-2, relative=True)

        # set parameters that need updating every frame
        # do any params need updating? (this method inherited from _base)
        if self.checkNeedToUpdate('set every frame'):
            buff.writeIndented(f"if {params['name']}.status == STARTED:  # only update if drawing\n")
            buff.setIndentLevel(+1, relative=True)  # to enter the if block
            self.writeParamUpdates(buff, 'set every frame')
            buff.setIndentLevel(-1, relative=True)  # to exit the if block

    def writeFrameCodeJS(self, buff):
        """Write the code that will be called every frame
        """
        params = self.params
        if "PsychoJS" not in self.targets:
            buff.writeIndented(f"// *{params['name']}* not supported by PsychoJS\n")
            return

        buff.writeIndentedLines(f"\n// *{params['name']}* updates\n")
        # writes an if statement to determine whether to draw etc
        self.writeStartTestCodeJS(buff)
        buff.writeIndented(f"{params['name']}.setAutoDraw(true);\n")
        # to get out of the if statement
        buff.setIndentLevel(-1, relative=True)
        buff.writeIndented("}\n\n")

        # test for stop (only if there was some setting for duration or stop)
        if self.params['stopVal'].val not in ('', None, -1, 'None'):
            # writes an if statement to determine whether to draw etc
            self.writeStopTestCodeJS(buff)
            buff.writeIndented(f"{params['name']}.setAutoDraw(false);\n")
            # to get out of the if statement
            buff.setIndentLevel(-1, relative=True)
            buff.writeIndented("}\n")

        # set parameters that need updating every frame
        # do any params need updating? (this method inherited from _base)
        if self.checkNeedToUpdate('set every frame'):
            buff.writeIndentedLines(f"\nif ({params['name']}.status === PsychoJS.Status.STARTED){{ "
                                    f"// only update if being drawn\n")
            buff.setIndentLevel(+1, relative=True)  # to enter the if block
            self.writeParamUpdatesJS(buff, 'set every frame')
            buff.setIndentLevel(-1, relative=True)  # to exit the if block
            buff.writeIndented("}\n")


def canBeNumeric(inStr):
    """Determines whether the input can be converted to a float
    (using a try: float(instr))
    """
    try:
        float(inStr)
        return True
    except Exception:
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo creating a drifting visual plaid stimulus.

For combining stimuli (e.g. to create a plaid) it's best to use blendMode='add'
rather than blendMode='avg'. In this blendMode the background is not overwritten
but added to, which is ideal in this instance.
On the other hand, in this mode the opacity attribute is a slight
misnomer; setting a high 'opacity' doesn't cause the background to be
obscured; it just acts as a multiplier for the contrast of the stimulus being drawn.
"""

from psychopy import visual, logging, event, core

# create a window to draw in
win = visual.Window((600, 600), allowGUI=False, blendMode='avg', useFBO=True)
logging.console.setLevel(logging.DEBUG)

# Initialize some stimuli, note contrast, opacity, ori
grating1 = visual.GratingStim(win, mask="circle", color='white', contrast=0.5,
    size=(1.0, 1.0), sf=(4, 0), ori = 45, autoLog=False)
grating2 = visual.GratingStim(win, mask="circle", color='white', opacity=0.5,
    size=(1.0, 1.0), sf=(4, 0), ori = -45, autoLog=False,
    pos=(0.1,0.1))

trialClock = core.Clock()
t = 0
while not event.getKeys() and t < 20:
    t = trialClock.getTime()

    grating1.phase = 1 * t  # drift at 1Hz
    grating1.draw()  # redraw it
    grating2.phase = 2 * t    # drift at 2Hz
    grating2.draw()  # redraw it

    win.flip()

win.close()
core.quit()

# The contents of this file are in the public domain.



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This demo shows you how to schedule a call to an arbitrary function immediately
using win.callOnFlip()

This is especially useful in scheduling a call to a trigger (eg. parallel port) when
your frame flip occurs. Avoid do anything time-consuming in these function calls.

The first argument to callOnFlip() is the function you want to call, the other
arguments are the arguments exactly as you would normally use them
(and can used ordered arguments or named, keyword args as usual).
"""

from psychopy import visual, core
import numpy

win = visual.Window([400, 400])
# insert a pause to allow the window and python (avoid initial frame drops)
core.wait(2)
clock = core.Clock()  # a clock to check times from

# a function to be called on selected frames
def printFrame(frameN, tReceived):
    tPrinted = clock.getTime()
    print(frameN, tReceived, tPrinted)

for frame in range(100):
    core.wait(numpy.random.random()/200)  # wait 0-5ms
    if frame in [2, 4, 6, 20, 30, 40]:
        win.callOnFlip(printFrame, frame, tReceived=clock.getTime())
    core.wait(numpy.random.random() / 200)  # wait 0-5ms
    win.flip()

win.close()
core.quit()

# The contents of this file are in the public domain.



# Provisional (super simple) test file which doesn't really test anything else
# besides simply opening a `pyGLFW` window. Add more complete `glfw` backend
# tests to other scripts ASAP, then remove this file.
from psychopy.tests import skip_under_vm
import pytest


@skip_under_vm(reason="GLFW doesn't work on (macOS) virtual machine")
def test_open_glfw_window():
    from psychopy.visual.window import Window
    win = Window(winType='glfw', autoLog=False)
    assert win.winType == 'glfw'
    win.flip()
    win.close()



"""Plugin manager for PsychoPy GUI apps (Builder and Coder)."""

from pkg_resources import parse_version
import wx

try:
    from wx import aui
except ImportError:
    import wx.lib.agw.aui as aui  # some versions of phoenix
try:
    from wx.adv import PseudoDC
except ImportError:
    from wx import PseudoDC

if parse_version(wx.__version__) < parse_version('4.0.3'):
    wx.NewIdRef = wx.NewId

from psychopy import plugins
from wx.lib.mixins.listctrl import ListCtrlAutoWidthMixin, CheckListCtrlMixin

from psychopy.preferences import prefs

import os


# Get a copy of startup plugins, we want to defer changes made to preferences to
# take effect after PsychoPy is shutdown. This prevents any sub-processed
# spawned by the GUI from using the plugins until a full restart.
if 'startUpPlugins' in prefs.general.keys():
    _startup_plugins_ = list(prefs.general['startUpPlugins'])
else:
    _startup_plugins_ = []

_startUpPluginsUpdated = False  # flag if plugins have been changed


class PluginBrowserListCtrl(wx.ListCtrl, ListCtrlAutoWidthMixin, CheckListCtrlMixin):
    """Custom ListCtrl that allows for automatic resizing of columns and
    checkboxes."""
    def __init__(self, parent, id):
        wx.ListCtrl.__init__(self,
                             parent,
                             id,
                             style=wx.LC_REPORT | wx.LC_SINGLE_SEL)
        ListCtrlAutoWidthMixin.__init__(self)
        CheckListCtrlMixin.__init__(self)

        # colors for rows
        colordb = wx.ColourDatabase()
        self.defaultRowColor = self.GetBackgroundColour()
        self.attnRowColor = colordb.Find('MEDIUM GOLDENROD')
        self.failedRowColor = colordb.Find('PLUM')

        self.pluginRemovedFlag = False

        self.createColumns()

    @property
    def selectedItem(self):
        return self.GetFirstSelected()

    def updatePluginStatus(self):
        """Update the plugin status column text and set the row color if a
        restart is needed.
        """
        for itemIdx in range(0, self.GetItemCount()):
            pluginName = self.GetItem(itemIdx, col=0).GetText()

            # deal with failed plugins
            if pluginName in plugins._failed_plugins_:
                if pluginName in _startup_plugins_:
                    self.SetItemBackgroundColour(
                        itemIdx, self.failedRowColor)
                    self.SetItem(itemIdx, 1, 'Failed')
                else:
                    self.SetItemBackgroundColour(
                        itemIdx, self.attnRowColor)
                    self.SetItem(itemIdx, 1, 'Needs Restart')

                continue

            if pluginName in _startup_plugins_:
                if not plugins.isPluginLoaded(pluginName):
                    self.SetItemBackgroundColour(
                        itemIdx, self.attnRowColor)
                    status = 'Needs Restart'
                else:
                    self.SetItemBackgroundColour(
                        itemIdx, self.defaultRowColor)
                    status = 'Ready'
            else:
                if plugins.isPluginLoaded(pluginName):
                    self.SetItemBackgroundColour(
                        itemIdx, self.attnRowColor)
                    status = 'Needs Restart'
                else:
                    self.SetItemBackgroundColour(
                        itemIdx, self.defaultRowColor)
                    status = ''

            self.SetItem(itemIdx, 1, status)


    def createColumns(self):
        """Create columns for this widget."""
        self.InsertColumn(0, 'Name', width=180)
        self.InsertColumn(1, 'Status', wx.LIST_FORMAT_CENTER, width=100)
        self.InsertColumn(2, 'Version', width=60)
        self.InsertColumn(3, 'Author', width=150)
        self.InsertColumn(4, 'Description', width=250)

    def refreshList(self):
        """Refresh the plugin list.
        """
        global _startup_plugins_
        self.DeleteAllItems()  # clear existing items

        # populate the list with installed plugins
        plugins.scanPlugins()
        allPlugins = plugins.listPlugins(which='all')

        # check if there are startup plugins that have been uninstalled during
        # the session
        for pluginName in _startup_plugins_:
            if pluginName not in allPlugins:
                # show the warning
                dlg = wx.MessageDialog(
                    self,
                    "Startup plugin `{}` cannot be found on the system! It "
                    "will be removed from startup plugins.".format(pluginName),
                    caption="Warning",
                    style=wx.OK | wx.CENTRE | wx.ICON_WARNING)
                dlg.ShowModal()

                # remove the startup plugin entry
                try:
                    _startup_plugins_.remove(pluginName)
                except KeyError:
                    pass

                self.pluginRemovedFlag = True

        # populate the list with installed plugins
        for pluginName in allPlugins:
            # get the metadata from the plugin to display
            metadata = plugins.pluginMetadata(pluginName)
            index = self.InsertItem(0, pluginName)

            # put a checkmark on the plugin if loaded
            self.CheckItem(index, pluginName in _startup_plugins_)
            self.SetItem(
                index, 2,
                metadata['Version'] if 'Version' in metadata.keys() else 'N/A')
            self.SetItem(
                index, 3,
                metadata['Author'] if 'Author' in metadata.keys() else 'N/A')
            self.SetItem(
                index, 4,
                metadata['Summary'] if 'Summary' in metadata.keys() else 'N/A')

        self.updatePluginStatus()

    def OnCheckItem(self, index, flag):
        """Do something when an item is checked."""
        item = self.GetItem(index, col=0)
        pluginName = item.GetText()
        global _startup_plugins_
        # check if in startup
        if not flag:
            try:
                _startup_plugins_.remove(pluginName)
            except ValueError:
                pass
        else:
            if pluginName not in _startup_plugins_:
                _startup_plugins_.append(pluginName)

        self.updatePluginStatus()

    def needsRestart(self):
        """Check if there are any items with status indicating a restart is
        needed."""
        self.refreshList()
        if self.pluginRemovedFlag:
            return True

        # if any items indicate a restart is needed, give a message
        for itemIdx in range(0, self.GetItemCount()):
            if self.GetItem(itemIdx, col=1).GetText() == "Needs Restart":
                return True

        return False

    def updateStartUpPluginsList(self):
        """Update the list of startup plugins based on what has been checked
        off in the list."""
        global _startup_plugins_

        # clear the startup plugins, only use the one from the manager
        _startup_plugins_ = []

        for itemIdx in range(0, self.GetItemCount()):
            if self.IsChecked(itemIdx):
                pluginName = self.GetItem(itemIdx, col=0).GetText()
                _startup_plugins_.append(pluginName)


class EntryPointListCtrl(wx.ListCtrl, ListCtrlAutoWidthMixin):
    """Custom ListCtrl that allows for automatic resizing of columns and
    checkboxes."""
    def __init__(self, parent, id):
        wx.ListCtrl.__init__(self,
                             parent,
                             id,
                             style=wx.LC_REPORT | wx.LC_SINGLE_SEL)
        ListCtrlAutoWidthMixin.__init__(self)


class EntryPointViewer(wx.Dialog):
    """Dialog that displays a plugin's entry points.
    """
    def __init__(self, parent, pluginName):
        """A dialog for loading and managing plugins.
        """
        self.parent = parent
        self.pluginName = pluginName
        title = "Plugin Entry Points"
        pos = wx.Point(parent.Position[0] + 80, parent.Position[1] + 80)
        wx.Dialog.__init__(self, parent, title=title,
                           size=(640, 380), pos=pos,
                           style=wx.DEFAULT_DIALOG_STYLE |
                                 wx.FRAME_FLOAT_ON_PARENT |
                                 wx.RESIZE_BORDER)
        self.Bind(wx.EVT_CLOSE, self.onClose)
        self.initCtrls()

    def onClose(self, evt=None):
        """
        Defines behavior on close of the Readme Frame
        """
        self.Destroy()

    def initCtrls(self):
        """Create window controls."""
        frameSizer = wx.BoxSizer(wx.VERTICAL)
        framePanel = wx.Panel(self)
        framePanelSizer = wx.BoxSizer(wx.VERTICAL)

        # add the box
        fraEntryPoints = wx.StaticBox(
            framePanel, wx.ID_ANY,
            "Entry points advertised by plugin `{}`".format(self.pluginName))
        fraSizer = wx.StaticBoxSizer(fraEntryPoints, wx.HORIZONTAL)

        lstEntryPoints = EntryPointListCtrl(fraEntryPoints, id=wx.ID_ANY)
        lstEntryPoints.InsertColumn(0, 'Group', width=200)
        lstEntryPoints.InsertColumn(1, 'Attribute', width=100)
        lstEntryPoints.InsertColumn(2, 'Entry Point', width=200)

        # populate the tree control
        entryPointMap = plugins.pluginEntryPoints(self.pluginName, parse=True)
        for group, entryPoints in entryPointMap.items():
            for attr, val in entryPoints.items():
                index = lstEntryPoints.InsertItem(0, group)
                lstEntryPoints.SetItem(index, 0, group)
                lstEntryPoints.SetItem(index, 1, attr)
                lstEntryPoints.SetItem(index, 2, val)

        fraSizer.Add(
            lstEntryPoints,
            flag=wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP | wx.BOTTOM,
            border=5,
            proportion=1)

        # panel for dialog buttons
        pnlDialogCtrls = wx.Panel(framePanel)
        pnlDialogCtrlsSizer = wx.FlexGridSizer(1, 1, 10, 10)
        pnlDialogCtrls.SetSizer(pnlDialogCtrlsSizer)

        # load selected plugin button
        self.cmdClose = wx.Button(pnlDialogCtrls, id=wx.ID_ANY, label='Close')
        self.cmdClose.Bind(wx.EVT_BUTTON, self.onClose)
        pnlDialogCtrlsSizer.Add(self.cmdClose, 0, 0)

        # add the panel to the frame sizer
        framePanelSizer.Add(
            fraSizer,
            flag=wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP | wx.BOTTOM,
            border=10,
            proportion=1)

        framePanelSizer.Add(
            pnlDialogCtrls,
            flag=wx.LEFT | wx.RIGHT | wx.BOTTOM | wx.ALIGN_RIGHT,
            border=10)

        # add the panel to the frame sizer

        framePanel.SetSizer(framePanelSizer)
        frameSizer.Add(framePanel, flag=wx.EXPAND | wx.ALL, proportion=1)
        self.SetSizer(frameSizer)


class PluginManagerFrame(wx.Dialog):
    """Defines the construction of the plugin manager frame.

    This provides a graphical interface for getting information about installed
    plugins and loading them into the current session. This UI is accessed
    through the "File" > "Plugin Manager" menu item in Builder and Coder.

    """
    def __init__(self, parent):
        """A frame for loading and managing plugins.
        """
        self.parent = parent
        title = "Plugins"
        pos = wx.Point(parent.Position[0] + 80, parent.Position[1] + 80)
        _style = wx.DEFAULT_DIALOG_STYLE | wx.FRAME_FLOAT_ON_PARENT | wx.RESIZE_BORDER
        wx.Dialog.__init__(self, parent, title=title,
                           size=(1024, 480), pos=pos, style=_style)
        self.Bind(wx.EVT_CLOSE, self.onClose)

        self.initCtrls()
        self.lstPlugins.refreshList()
        self.cmdEntryPoints.Disable()

    def initCtrls(self):
        """Create window controls."""
        frameSizer = wx.BoxSizer(wx.VERTICAL)
        framePanel = wx.Panel(self)
        panelSizer = wx.BoxSizer(wx.VERTICAL)

        pnlDialogHeader = wx.Panel(framePanel)
        pnlDialogHeaderSizer = wx.BoxSizer(wx.HORIZONTAL)

        # this should be cached ...
        PNG = wx.BITMAP_TYPE_PNG
        rc = prefs.paths['resources']
        pluginBMP = wx.Bitmap(os.path.join(rc, 'plugins32.png'), PNG)
        pluginGraphic = wx.StaticBitmap(
            pnlDialogHeader, wx.ID_ANY, pluginBMP, (0, 0), (32, 32))

        # add some text
        lblInfo = wx.StaticText(
            pnlDialogHeader,
            id=wx.ID_ANY,
            label="Plugins are third-party packages used to extend PsychoPy. "
                  "Indicate below which plugins should be loaded when a "
                  "PsychoPy session starts.")

        pnlDialogHeaderSizer.Add(pluginGraphic, flag=wx.ALIGN_CENTRE_VERTICAL)
        pnlDialogHeaderSizer.Add(
            lblInfo, flag=wx.ALIGN_CENTRE_VERTICAL | wx.LEFT, border=10)

        pnlDialogHeader.SetSizer(pnlDialogHeaderSizer)
        panelSizer.Add(
            pnlDialogHeader,
            flag=wx.EXPAND | wx.TOP | wx.LEFT | wx.RIGHT,
            border=10)

        # add the box
        fraPlugins = wx.StaticBox(framePanel, wx.ID_ANY, "Available Plugins")
        fraSizer = wx.StaticBoxSizer(fraPlugins, wx.HORIZONTAL)
        pnlPlugins = wx.Panel(fraPlugins)
        bsizer = wx.BoxSizer(wx.HORIZONTAL)

        # plugin list
        self.lstPlugins = PluginBrowserListCtrl(pnlPlugins, id=wx.ID_ANY)
        self.lstPlugins.Bind(wx.EVT_LIST_ITEM_SELECTED, self.onItemSelected)
        self.lstPlugins.Bind(wx.EVT_LIST_ITEM_DESELECTED, self.onItemSelected)

        bsizer.Add(self.lstPlugins, flag=wx.EXPAND | wx.ALL, proportion=1)
        fraSizer.Add(pnlPlugins, flag=wx.EXPAND | wx.ALL, proportion=1, border=5)

        # plugin buttons
        buttonSizer = wx.BoxSizer(wx.VERTICAL)

        # rescan plugins button
        self.cmdScanPlugin = wx.Button(pnlPlugins, id=wx.ID_ANY, label='Rescan')
        self.cmdScanPlugin.Bind(wx.EVT_BUTTON, self.onRescanPlugins)
        self.cmdScanPlugin.SetToolTip(wx.ToolTip(
            "Rescan installed packages for PsychoPy plugins."))
        buttonSizer.Add(
            self.cmdScanPlugin,
            flag=wx.EXPAND | wx.BOTTOM,
            border=5)

        # display entry points button
        self.cmdEntryPoints = wx.Button(pnlPlugins, id=wx.ID_ANY,
                                        label='Entry Points ...')
        self.cmdEntryPoints.Bind(wx.EVT_BUTTON, self.onShowEntryPoints)
        self.cmdEntryPoints.SetToolTip(wx.ToolTip(
            "Display the entry points for the selected plugin."))
        buttonSizer.Add(
            self.cmdEntryPoints,
            flag=wx.EXPAND | wx.BOTTOM,
            border=5)

        bsizer.Add(buttonSizer, flag=wx.EXPAND | wx.BOTTOM | wx.LEFT, border=5)
        pnlPlugins.SetSizer(bsizer)

        # panel for dialog buttons
        pnlDialogCtrls = wx.Panel(framePanel)
        pnlDialogCtrlsSizer = wx.FlexGridSizer(1, 2, 10, 10)
        pnlDialogCtrls.SetSizer(pnlDialogCtrlsSizer)

        # disable all startup plugins button
        self.cmdDisableAll = wx.Button(
            pnlDialogCtrls,
            id=wx.ID_ANY,
            label='Clear startup plugins')
        self.cmdDisableAll.Bind(wx.EVT_BUTTON, self.onClearStartupPlugins)
        self.cmdDisableAll.SetToolTip(wx.ToolTip(
            "Clear all plugins registered to load on startup."))
        pnlDialogCtrlsSizer.Add(self.cmdDisableAll, 0, 0)

        # load selected plugin button
        self.cmdClose = wx.Button(pnlDialogCtrls, id=wx.ID_ANY, label='Close')
        self.cmdClose.Bind(wx.EVT_BUTTON, self.onClose)
        pnlDialogCtrlsSizer.Add(self.cmdClose, 0, 0)

        # add the panel to the frame sizer
        panelSizer.Add(
            fraSizer,
            flag=wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP | wx.BOTTOM,
            border=10,
            proportion=1)
        panelSizer.Add(
            pnlDialogCtrls,
            flag=wx.LEFT | wx.RIGHT | wx.BOTTOM | wx.ALIGN_RIGHT,
            border=10)
        framePanel.SetSizer(panelSizer)
        frameSizer.Add(framePanel, flag=wx.EXPAND | wx.ALL, proportion=1)
        self.SetSizer(frameSizer)

    def onRescanPlugins(self, evt=None):
        """Event handler for when the rescan button is pressed."""
        self.cmdEntryPoints.Disable()
        self.lstPlugins.refreshList()

    def onShowEntryPoints(self, evt=None):
        """Event for when the entry points of a selected plugin are
        requested."""
        selectedItem = self.lstPlugins.selectedItem
        if selectedItem == -1:
            return

        # show the entry point dialog
        epView = EntryPointViewer(self, self.lstPlugins.GetItem(
            selectedItem, col=0).GetText())
        epView.ShowModal()

    def onClearStartupPlugins(self, evt=None):
        """Clear all startup plugins."""
        global _startup_plugins_
        _startup_plugins_ = []
        self.lstPlugins.refreshList()

    def onItemSelected(self, evt=None):
        """Event handler for when an item is selected."""
        self.selectedItem = self.lstPlugins.GetFirstSelected()
        if self.lstPlugins.selectedItem != -1:
            self.cmdEntryPoints.Enable()
        else:
            self.cmdEntryPoints.Disable()

    def onClose(self, evt=None):
        """Called when the plugin manager is closed.
        """
        global _startUpPluginsUpdated
        _startUpPluginsUpdated = self.lstPlugins.needsRestart()

        # warn if there are any plugins that need a restart
        if _startUpPluginsUpdated:
            dlg = wx.MessageDialog(
                self,
                "PsychoPy must be restarted for plugin changes to take effect.",
                caption="Information",
                style=wx.OK | wx.CENTRE | wx.ICON_INFORMATION)
            dlg.ShowModal()

            # Update the startup plugins based off what has been given in the
            # list. Only do this if the user made any changes.
            self.lstPlugins.updateStartUpPluginsList()

        self.parent.pluginManager = None
        self.Destroy()


def saveStartUpPluginsConfig():
    """Write startup plugins to the user config. This will only write to the
    config file if there have been changes to the plugin configuration over the
    course of the current session.

    """
    if _startUpPluginsUpdated:
        prefs.general['startUpPlugins'] = _startup_plugins_
        prefs.saveUserPrefs()



""" Test starting and stopping iohub server
"""
from psychopy.tests import skip_under_vm
from psychopy.tests.test_iohub.testutil import startHubProcess, stopHubProcess

@skip_under_vm
class TestKeyboard():
    """
    Keyboard Device tests. Starts iohub server, runs test set, then
    stops iohub server.

    Since there is no way to currently automate keyboard event generation in
    a way that would actually test the iohub keyboard event processing logic,
    each test simply calls one of the device methods / properties and checks
    that the return type is as expected.

    Each method is called with no args; that should be improved.

    Following methods are not yet tested:
            addFilter
            enableFilters
            getConfiguration
            getCurrentDeviceState
            getModifierState
            removeFilter
            resetFilter
            resetState
    """

    @classmethod
    def setup_class(cls):
        """ setup any state specific to the execution of the given class (which
        usually contains tests).
        """
        cls.io = startHubProcess()
        cls.keyboard = cls.io.devices.keyboard

    @classmethod
    def teardown_class(cls):
        """ teardown any state that was previously setup with a call to
        setup_class.
        """
        stopHubProcess()
        cls.io = None
        cls.keyboard = None

    def test_getEvents(self):
        evts = self.keyboard.getEvents()
        assert type(evts) in [list, tuple]

    def test_getKeys(self):
        evts = self.keyboard.getKeys()
        assert type(evts) in [list, tuple]

    def test_getPresses(self):
        evts = self.keyboard.getPresses()
        assert type(evts) in [list, tuple]

    def test_getReleases(self):
        evts = self.keyboard.getReleases()
        assert type(evts) in [list, tuple]

    def test_waitForKeys(self):
        evts = self.keyboard.waitForKeys(maxWait=0.05)
        assert type(evts) in [list, tuple]

    def test_waitForPresses(self):
        evts = self.keyboard.waitForPresses(maxWait=0.05)
        assert type(evts) in [list, tuple]

    def test_waitForReleases(self):
        evts = self.keyboard.waitForReleases(maxWait=0.05)
        assert type(evts) in [list, tuple]

    def test_clearEvents(self):
        self.keyboard.clearEvents()

    def test_state(self):
        kbstate = self.keyboard.state
        assert type(kbstate) is dict

    def test_reporting(self):
        reporting_state = self.keyboard.reporting
        assert reporting_state is True

        self.keyboard.reporting = False

        assert self.keyboard.isReportingEvents() is False

        self.keyboard.reporting = True
        assert self.keyboard.isReportingEvents() is True
